# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
import numpy as N
import hashlib


class NumpyUtil(object):
    """
    Numpy ndarray object utility abstract class
    """

    @classmethod
    def check_is_array(cls, a, ndim=None):
        """
        Checks that an object is a valid numpy.ndarray object, optionally with a given number of dimensions

        Parameters
        ----------
        a: :obj:`numpy.ndarray`
            array to check
        ndim: :obj:`int`` or None
            number of dimensions to check, if not None
        """
        if not isinstance(a, N.ndarray):
            raise AttributeError("'a' is not a valid numpy.ndarray object")

        if ndim is not None:
            if a.ndim != ndim:
                raise AttributeError("'a' does not have {n!s} dimensions.".format(n=ndim))

    @classmethod
    def check_dtype(cls, a):
        """
        Check numpy.ndarray datatype

        Parameters
        ----------
        a: :obj:`numpy.ndarray`
            array

        Returns
        -------
        k: :obj:`string`
            array datatype kind
        """
        dtk = a.dtype.kind
        if dtk in ["U", "S"]:
            return "s"
        elif dtk == "f":
            return "f"
        elif dtk == "i":
            return "i"
        elif dtk == 'b':
            return "b"
        return None

    @classmethod
    def md5sum(cls, a):
        """
        Compute the md5 checksum of a given array

        Parameter
        ---------
        a: `numpy.ndarray`
            Numpy arrray
        """
        cls.check_is_array(a)
        hash_md5 = hashlib.md5()
        hash_md5.update(a)
        return hash_md5.hexdigest()


__all__ = ["NumpyUtil"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import absolute_import, unicode_literals
import uuid
from future.builtins import str, list, dict
import logging

from astrophysix.simdm.catalogs import CatalogField
from astrophysix.simdm.utils import GalacticaValidityCheckMixin, ObjectList, DataType
from astrophysix.utils.persistency import Hdf5StudyPersistent
from astrophysix.utils.strings import Stringifiable

log = logging.getLogger("astrophysix.simdm")


class DataProcessingService(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Data processing service description defining a service name and a data host server name. Use it to bind
    a Snapshot to a specific data-processing service on Galactica.

    Parameters
    ----------
    service_name: :obj:`string`
        data processing service name (mandatory)
    data_host: :obj:`string`
        data host server name name (mandatory)

    Example
    -------

        >>> # To bind a given simulation snapshot to a data-processing service :
        >>> sn = Snapshot(name="Third pericenter", time=(254.7, U.Myr),
        ...               data_reference="output_00034")
        >>> serv = DataProcessingService(service_name="ray_tracer_amr",
        ...                              data_host="my_institute_cluster")
        >>> sn.processing_services.add(serv)

    """
    def __init__(self, **kwargs):
        super(DataProcessingService, self).__init__(**kwargs)

        self._service_name = ""
        self._data_host = ""

        if "service_name" not in kwargs:
            raise AttributeError("Data processing service 'service_name' attribute is not defined (mandatory).")
        self.service_name = kwargs["service_name"]

        if "data_host" not in kwargs:
            raise AttributeError("Data processing service 'data_host' attribute is not defined (mandatory).")
        self.data_host = kwargs["data_host"]

    def __eq__(self, other):
        """
        DataProcessingService comparison method

        other: :class:`~astrophysix.simdm.services.DataProcessingService`
            Data processing service to compare to
        """
        if not super(DataProcessingService, self).__eq__(other):
            return False

        if self._service_name != other.service_name:
            return False

        if self._data_host != other.data_host:
            return False

        return True

    @property
    def service_name(self):
        """Data processing service name"""
        return self._service_name

    @service_name.setter
    def service_name(self, new_service_name):
        try:
            self._service_name = Stringifiable.cast_string(new_service_name, valid_empty=False)
        except TypeError:
            raise AttributeError("Data processing service 'service_name' property is not a valid (non empty) string.")

    @property
    def data_host(self):
        """Data processing service host server name"""
        return self._data_host

    @data_host.setter
    def data_host(self, new_host):
        try:
            self._data_host = Stringifiable.cast_string(new_host, valid_empty=False)
        except TypeError:
            raise AttributeError("Data processing service 'data_host' property is not a valid (non empty) string.")

    @property
    def hosted_service(self):
        """Data processing service full description *service_name @ data_host*"""
        return "{sn:s} @ {hn:s}".format(sn=self._service_name, hn=self._data_host)

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["service_name", "data_host"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize an DataProcessingService object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the DataProcessingService into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(DataProcessingService, self)._hsp_write(h5group, **kwargs)

        # Write data processing service name
        self._hsp_write_attribute(h5group, ('service_name', self._service_name), **kwargs)

        # Write data processing host server name
        self._hsp_write_attribute(h5group, ('data_host', self._data_host), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read an DataProcessingService object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        serv: ``DataProcessingService``
            Read DataProcessingService instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(DataProcessingService, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read data processing service name
        serv_name = cls._hsp_read_attribute(h5group, 'service_name', "data processing service name")

        # Read data processing host server name
        host_name = cls._hsp_read_attribute(h5group, 'data_host', "data processing host server name")

        # Create data processing service object
        serv = cls(service_name=serv_name, data_host=host_name, uid=uid)

        return serv

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check data host + service name lengths
        if len(self._data_host) > 16:
            log.warning("Data processing service host name '{h:s}' is too long for Galactica (max. 16 "
                        "characters).".format(h=self._data_host))
        if len(self._service_name) > 32:
            log.warning("Data processing service name '{s:s}' is too long for Galactica (max. 32 "
                        "characters).".format(s=self._service_name))

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{sn:s}' data processing service (host: {hn:s})".format(sn=self._service_name, hn=self._data_host)


class CatalogFieldBinding(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Service input parameter - catalog field value binding for (catalog-bound) data processing services. The applied
    scaling formula is : :math:`\\textrm{param_value} = \\textrm{scale} \\times \\textrm{field_value} + \\textrm{offset}`.

    Parameters
    ----------
    catalog_field: :class:`~astrophysix.simdm.catalogs.field.CatalogField`
        bound catalog field (mandatory)
    param_key: :obj:`string`
        data processing service input parameter key (mandatory)
    scale: :obj:`float`
        field value to service parameter scaling factor. Default 1.0
    offset: : obj:`float`
        field value to service parameter offset value. Default 0.0

    Example
    -------

        >>> # Define a catalog
        >>> cat = Catalog(target_object=gal_cluster, name="Galaxy cluster catalog")
        >>> # Add the catalog fields into the catalog (100 clusters)
        >>> fx = cat.catalog_fields.add(CatalogField(x, values=N.random.uniform(size=100)))
        >>> fy = cat.catalog_fields.add(CatalogField(y, values=N.random.uniform(size=100)))
        >>> fz = cat.catalog_fields.add(CatalogField(z, values=N.random.uniform(size=100)))
        >>> fm = cat.catalog_fields.add(CatalogField(m, values=N.random.uniform(size=100)))
        >>> # To bind a given object catalog to a data-processing service :
        >>> cat_dps = CatalogDataProcessingService(service_name="slice_map",
        ...                                     data_host="Lab_Cluster")
        >>> cat.processing_services.add(cat_dps)
        >>>
        >>> # Define catalog field bindings to automatically fill the data processing service
        >>> # parameter value 'pv' with a catalog field value 'fv' of one of your catalog's object
        >>> # according to the formula : pv = scale * fv + offset.
        >>> fbx = CatalogFieldBinding(catalog_field=fx, param_key="x", scale=1.0e2, offset=-50.0))
        >>> fby = CatalogFieldBinding(catalog_field=fy, param_key="y", scale=1.0e2, offset=-50.0))
        >>> fbz = CatalogFieldBinding(catalog_field=fz, param_key="z", scale=1.0e2, offset=-50.0))
        >>> cat_dps.catalog_field_bindings.add(fbx)
        >>> cat_dps.catalog_field_bindings.add(fby)
        >>> cat_dps.catalog_field_bindings.add(fbz)

    """
    def __init__(self, *args, **kwargs):
        super(CatalogFieldBinding, self).__init__(**kwargs)

        self._param_key = ""

        # Bound catalog field
        self._bound_field = None
        if len(args) > 0:
            cat_field = args[0]
        elif "catalog_field" in kwargs:
            cat_field = kwargs["catalog_field"]
        else:
            raise AttributeError("Undefined 'catalog_field' attribute in "
                                 "{cname:s}.".format(cname=self.__class__.__name__))

        if not isinstance(cat_field, CatalogField):
            err_msg = "{cname:s} 'catalog_field' attribute is not a valid CatalogField " \
                      "instance.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._bound_field = cat_field

        # Service parameter key
        if "param_key" not in kwargs:
            raise AttributeError("{cname:s} 'param_key' attribute is not defined "
                                 "(mandatory).".format(cname=self.__class__.__name__))
        self.param_key = kwargs["param_key"]

        # Init scale/factor
        self._scale_factor = 1.0
        if "scale" in kwargs:
            self.scale = kwargs["scale"]
        self._offset = 0.0
        if "offset" in kwargs:
            self.offset = kwargs["offset"]

    def __eq__(self, other):
        """
        CatalogDataProcessingService comparison method

        other: :class:`~astrophysix.simdm.services.process.CatalogFieldBinding`
            catalog field binding to compare to
        """
        if not super(CatalogFieldBinding, self).__eq__(other):
            return False

        # Compare service parameter keys
        if self._param_key != other.param_key:
            return False

        # Compare catalog fields
        if self._bound_field != other.catalog_field:
            return False

        # Compare scale/offset
        if self._scale_factor != other.scale or self._offset != other.offset:
            return False

        return True

    @property
    def field_property_name(self):
        """Associated catalog field's target object property name. Cannot be edited"""
        return self._bound_field.property_name

    @property
    def catalog_field(self):
        """Associated catalog field (:class:`~astrophysix.simdm.catalogs.field.CatalogField`). Cannot be edited."""
        return self._bound_field

    @property
    def param_key(self):
        """Data processing service parameter key. Can be edited."""
        return self._param_key

    @param_key.setter
    def param_key(self, new_param_key):
        try:
            self._param_key = Stringifiable.cast_string(new_param_key, valid_empty=False)
        except TypeError:
            err_msg = "{cname:s} 'param_key' property is not a valid (non-empty) " \
                      "string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def scale(self):
        """
        Returns the scaling factor to apply between the catalog field value and input value provided to the
        data-processing service parameter value. Can be edited.
        """
        return self._scale_factor

    @scale.setter
    def scale(self, new_scale):
        try:
            self._scale_factor = float(new_scale)
        except (ValueError, TypeError):  # Not a numeric/string value that can be casted into a real number
            raise AttributeError("{cn:s} 'scale' attribute is not a valid numeric "
                                 "value.".format(cn=self.__class__.__name__))

    @property
    def offset(self):
        """
        Returns the offset to apply between the catalog field value and input value provided to the data-processing
        service parameter value. Can be edited.
        """
        return self._offset

    @offset.setter
    def offset(self, new_offset):
        try:
            self._offset = float(new_offset)
        except (ValueError, TypeError):  # Not a numeric/string value that can be casted into a real number
            raise AttributeError("{cn:s} 'offset' attribute is not a valid numeric "
                                 "value.".format(cn=self.__class__.__name__))

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["catalog_field", "param_key", "scale", "offset"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize an CatalogFieldBinding object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the CatalogFieldBinding into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(CatalogFieldBinding, self)._hsp_write(h5group, **kwargs)

        # Write catalog field UUID
        self._hsp_write_attribute(h5group, ('cat_field_uid', self._bound_field.uid), **kwargs)

        # Write data processing service parameter key
        self._hsp_write_attribute(h5group, ('serv_param_key', self._param_key), **kwargs)

        # Write scale/offset parameters
        self._hsp_write_attribute(h5group, ('scale_fact', self._scale_factor), **kwargs)
        self._hsp_write_attribute(h5group, ('offset', self._offset), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read an CatalogFieldBinding object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        serv: ``CatalogFieldBinding``
            Read CatalogFieldBinding instance
        """
        # Handle different versions here

        # Read catalog field binding
        uid = super(CatalogFieldBinding, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read data processing service parameter key
        param_key = cls._hsp_read_attribute(h5group, 'serv_param_key', "service parameter key")

        # Try to read/find protocol
        try:
            cat_field_uid = uuid.UUID(cls._hsp_read_attribute(h5group, "cat_field_uid", "catalog field UUID",
                                                              raise_error_if_not_found=True))

            # Search for already instantiated catalog field in dependency object dictionary
            if dependency_objdict is None:
                err_msg = "Cannot find any catalog field object already instantiated in the project."
                log.error(err_msg)
                raise IOError(err_msg)

            # Get dictionary of catalog field instances of the corresponding class :
            if CatalogField.__name__ not in dependency_objdict:
                err_msg = "Cannot find any {cname:s} instance.".format(cname=CatalogField.__name__)
                log.error(err_msg)
                raise IOError(err_msg)

            # Find catalog field according to its UUID
            catfield_dict = dependency_objdict[CatalogField.__name__]
            if cat_field_uid not in catfield_dict:
                err_msg = "Cannot find {cname:s} instance with uid {uid:s}.".format(cname=CatalogField.__name__,
                                                                                    uid=str(uid))
                log.error(err_msg)
                raise IOError(err_msg)

            cfield = catfield_dict[cat_field_uid]
        except IOError:  # Catalog field UUID not found in CatalogFieldBinding
            raise

        binding = cls(uid=uid, param_key=param_key, catalog_field=cfield)

        # Read scale/offset attributes
        binding.scale = cls._hsp_read_attribute(h5group, 'scale_fact', "binding scaling factor")
        binding.offset = cls._hsp_read_attribute(h5group, 'offset', "binding offset")

        return binding

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check data-processing service parameter name length
        if len(self._param_key) > 16:
            log.warning("Data processing service parameter key '{p:s}' is too long for Galactica (max. 16 "
                        "characters).".format(p=self._param_key))

        # Check target object property type is numerical
        if self._bound_field.object_property.datatype not in [DataType.REAL, DataType.INTEGER]:
            log.warning("{cfb!s} is not a numerical type, it cannot be bound to a data processing service "
                        "parameter.".format(cfb=self))

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{fn:s}' catalog field binding".format(fn=self.field_property_name)


class CatalogDataProcessingService(DataProcessingService):
    """
    Catalog item-bound data-processing service defining a service name and a data host server name. Use it to bind
    a Catalog to a specific data-processing service on Galactica.

    Parameters
    ----------
    service_name: :obj:`string`
        data processing service name (mandatory)
    data_host: :obj:`string`
        data host server name (mandatory)

    Example
    -------

        >>> # Define a catalog
        >>> cat = Catalog(target_object=gal_cluster, name="Galaxy cluster catalog")
        >>>
        >>> # To bind a given object catalog to a data-processing service :
        >>> cat_dps = CatalogDataProcessingService(service_name="slice_map",
        ...                                        data_host="Lab_Cluster")
        >>> cat.processing_services.add(cat_dps)
    """
    def __init__(self, **kwargs):
        super(CatalogDataProcessingService, self).__init__(**kwargs)
        self._cat_field_bindings = ObjectList(CatalogFieldBinding, "param_key")

    def __eq__(self, other):
        """
        CatalogDataProcessingService comparison method

        other: :class:`~astrophysix.simdm.services.CatalogDataProcessingService`
            Data processing service to compare to
        """
        if not super(CatalogDataProcessingService, self).__eq__(other):
            return False

        if self._cat_field_bindings != other.catalog_field_bindings:
            return False

        return True

    @property
    def catalog_field_bindings(self):
        """Catalog data processing service list of catalog field bindings"""
        return self._cat_field_bindings

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize an CatalogDataProcessingService object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the CatalogDataProcessingService into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, data host, service name, etc.
        super(CatalogDataProcessingService, self)._hsp_write(h5group, **kwargs)

        # Write catalog field bindings, if any defined
        self._hsp_write_object_list(h5group, "CAT_FIELD_BINDINGS", self._cat_field_bindings, "catfbind_", **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read an CatalogDataProcessingService object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        serv: ``CatalogDataProcessingService``
            Read CatalogDataProcessingService instance
        """
        # Handle different versions here

        # Read data processing service
        service = super(CatalogDataProcessingService, cls)._hsp_read(h5group, version,
                                                                     dependency_objdict=dependency_objdict)

        # Build catalog field binding list and add each catalog field binding into service, if any defined
        if "CAT_FIELD_BINDINGS" in h5group:
            for cfb in CatalogFieldBinding._hsp_read_object_list(h5group, "CAT_FIELD_BINDINGS", "catfbind_",
                                                                 "catalog field binding",
                                                                 dependency_objdict=dependency_objdict):
                service.catalog_field_bindings.add(cfb)

        return service

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        super(CatalogDataProcessingService, self).galactica_validity_check(**kwargs)

        # Check catalog field bind list validity
        self._cat_field_bindings.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{sn:s}' catalog-bound data processing service (host: {hn:s})".format(sn=self.service_name,
                                                                                      hn=self.data_host)


__all__ = ["DataProcessingService", "CatalogDataProcessingService", "CatalogFieldBinding"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, list, int
import logging

from .generic import GenericResult
from astrophysix.utils.strings import Stringifiable
from astrophysix import units as U
from ..services import DataProcessingService
from ..utils import ObjectList

log = logging.getLogger("astrophysix.simdm")


class Snapshot(GenericResult):
    """
    Experiment snapshot class (Simulation data model)

    Parameters
    ----------
    name: :obj:`string`
        snapshot name (mandatory)
    description: :obj:`string`
        snapshot description
    directory_path: :obj:`string`
        snapshot directory path
    time: (:obj:`float`, :class:`~astrophysix.units.unit.Unit`) :obj:`tuple`
        snapshot time info (value, unit) tuple
    physical_size: (:obj:`float`, :class:`~astrophysix.units.unit.Unit`) :obj:`tuple`
        snapshot physical size info (value, unit) tuple
    data_reference: :obj:`string`
        snapshot data reference (e.g. data directory name, snapshot number) string
    """
    _hsp_version = 2  # With data processing services
    def __init__(self, **kwargs):
        super(Snapshot, self).__init__(**kwargs)
        self._time = 0.0
        self._time_unit = U.none
        self._phys_size = 0.0
        self._phys_size_unit = U.none
        self._data_reference = ""
        self._proc_services = ObjectList(DataProcessingService, "hosted_service")

        if "time" in kwargs:
            self.time = kwargs["time"]

        if "physical_size" in kwargs:
            self.physical_size = kwargs["physical_size"]

        if "data_reference" in kwargs:
            self.data_reference = kwargs["data_reference"]

    def __eq__(self, other):
        """
        Snapshot comparison method

        other: :class:`~astrophysix.simdm.results.snapshot.Snapshot`
            snapshot to compare to
        """
        if not super(Snapshot, self).__eq__(other):
            return False

        otime, otime_unit = other.time
        if self._time != otime or self._time_unit != otime_unit or self._time_unit.name != otime_unit.name:
            return False

        opsize, opsize_unit = other.physical_size
        if self._phys_size != opsize or self._phys_size_unit != opsize_unit or \
                self._phys_size_unit.name != opsize_unit.name:
            return False

        if self._data_reference != other.data_reference:
            return False

        if self._proc_services != other.processing_services:
            return False

        return True

    @property
    def time(self):
        """Snapshot time info (value, unit) tuple . Can be set to a :obj:`float` value (unitless) or a
        (:obj:`float`, :class:`~astrophysix.units.unit.Unit`) tuple.

        Example
        -------
            >>> sn = Snapshot(name="My super snapshot")
            >>> sn.time = "0.256"
            >>> sn.time[1] == U.none
            True
            >>> sn.time = ("0.24", U.year)
            >>> sn.time = ("0.45", "Myr")
            >>> sn.time[1] == U.Myr
            True
            >>> sn.time = (7.89e2, "Gyr")
            >>> sn.time = (78.54, U.min)
        """
        return self._time, self._time_unit

    @time.setter
    def time(self, new_time):
        if isinstance(new_time, tuple) or isinstance(new_time, list):
            if len(new_time) != 2:
                err_msg = "Snapshot 'time' property cannot be of length != 2"
                log.error(err_msg)
                raise AttributeError(err_msg)
            t_val, t_unit = new_time

            if not isinstance(t_unit, U.Unit):
                try:
                    s = Stringifiable.cast_string(t_unit, valid_empty=False)
                    t_unit = U.Unit.from_name(s)
                except TypeError:  # Not a valid string
                    err_msg = "Snapshot 'time' property must be defined with a valid (non-empty) time unit string."
                    log.error(err_msg)
                    raise AttributeError(err_msg)
                except AttributeError as aerr:
                    err_msg = "Snapshot 'time' property error : {uerr:s}.".format(uerr=str(aerr))
                    log.error(err_msg)
                    raise AttributeError(err_msg)

            if t_unit.physical_type not in ["time", "dimensionless"]:
                err_msg = "Error while setting {cn:s} 'time' property : unit is not a valid time unit (physical " \
                          "type: '{pt:s}')".format(pt=t_unit.physical_type, cn=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)

            self._time_unit = t_unit
        else:
            self._time_unit = U.none
            t_val = new_time

        try:
            self._time = float(t_val)
        except ValueError:
            err_msg = "Snapshot 'time' property must be set as a (time_float_value, time_unit) tuple."
            log.error(err_msg)
            raise AttributeError(err_msg)
        except TypeError:
            err_msg = "Snapshot 'time' property must be set as a (time_float_value, time_unit) tuple."
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def physical_size(self):
        """
        Snapshot physical size info (value, unit) tuple . Can be set to a :obj:`float` value (unitless) or a
        (:obj:`float`, :class:`~astrophysix.units.unit.Unit`) tuple.

        Example
        -------
            >>> sn = Snapshot(name="My super snapshot")
            >>> sn.physical_size = "0.256"
            >>> sn.physical_size = ("0.24", U.pc)
            >>> sn.physical_size = ("0.45", "kpc")
            >>> sn.physical_size[1] == U.kpc
            True
            >>> sn.physical_size = 4.46
            >>> sn.physical_size = (7.89e2, "Mpc")
            >>> sn.physical_size[1] == U.Mpc
            True
            >>> sn.physical_size = (78.54, U.ly)
        """
        return self._phys_size, self._phys_size_unit

    @physical_size.setter
    def physical_size(self, new_psize):
        if isinstance(new_psize, tuple) or isinstance(new_psize, list):
            if len(new_psize) != 2:
                err_msg = "Snapshot 'physical_size' property cannot be of length != 2"
                log.error(err_msg)
                raise AttributeError(err_msg)
            s_val, s_unit = new_psize

            if not isinstance(s_unit, U.Unit):
                try:
                    s = Stringifiable.cast_string(s_unit, valid_empty=False)
                    s_unit = U.Unit.from_name(s)
                except TypeError:  # Not a valid string
                    err_msg = "Snapshot 'physical_size' property must be defined with a valid (non-empty) length " \
                              "unit string."
                    log.error(err_msg)
                    raise AttributeError(err_msg)
                except AttributeError as aerr:
                    err_msg = "Snapshot 'physical_size' property error : {uerr:s}.".format(uerr=str(aerr))
                    log.error(err_msg)
                    raise AttributeError(err_msg)

            if s_unit.physical_type not in ["length", "dimensionless"]:
                err_msg = "Error while setting Snaphsot 'physical_size' property : unit is not a valid length unit " \
                          "(physical type: '{pt:s}')".format(pt=s_unit.physical_type)
                log.error(err_msg)
                raise AttributeError(err_msg)

            self._phys_size_unit = s_unit
        else:
            self._phys_size_unit = U.none
            s_val = new_psize

        try:
            self._phys_size = float(s_val)
        except ValueError:
            err_msg = "Snapshot 'physical_size' property must be set as a (size_float_value, length_unit) tuple."
            log.error(err_msg)
            raise AttributeError(err_msg)
        except TypeError:
            err_msg = "Snapshot 'physical_size' property must be set as a (size_float_value, length_unit) tuple."
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def data_reference(self):
        """Snapshot data reference (e.g. data directory name, snapshot number). Can be set to any :obj:`string` value."""
        return self._data_reference

    @data_reference.setter
    def data_reference(self, new_data_ref):
        try:
            self._data_reference = Stringifiable.cast_string(new_data_ref, valid_empty=True)
        except TypeError:
            err_msg = "{cname:s} 'data_reference' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def processing_services(self):
        """:class:`~astrophysix.simdm.services.process.DataProcessingService` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._proc_services

    def _can_add_catalog(self, new_cat):
        """
        Catalog addition validity check method. Overrides GenericResult._can_add_catalog()

        Parameters
        ----------
        new_cat: :class:`~astrophysix.simdm.catalogs.catalog.Catalog`
            new Catalog to add into the Snapshot
        """
        pass

    def _can_add_data_proc_service_in_catalog(self, cat_data_proc_service):
        """
        CatalogDataProcessingService addition validity check method. Verifies that the added catalog data processing
        service is added into a Snapshot's catalog.

        Parameters
        ----------
        cat_data_proc_service: :class:`~astrophysix.simdm.services.process.CatalogDataProcessingService`
            new catalog <=> data processing service to add into this Catalog.
        """
        pass

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        genres_valid_attrs = super(Snapshot, cls)._hsp_valid_attributes()
        genres_valid_attrs += ["time", "physical_size", "data_reference"]
        return genres_valid_attrs

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a Snapshot object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the Snapshot into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write()
        super(Snapshot, self)._hsp_write(h5group, **kwargs)

        # Write snapshot time
        self._hsp_write_attribute(h5group, ('time', self._time), **kwargs)
        self._hsp_write_attribute(h5group, ('time_unit', self._time_unit), **kwargs)

        # Write snapshot physical size
        self._hsp_write_attribute(h5group, ('phys_size', self._phys_size), **kwargs)
        self._hsp_write_attribute(h5group, ('phys_size_unit', self._phys_size_unit), **kwargs)

        # Write snapshot data reference
        self._hsp_write_attribute(h5group, ("data_reference", self._data_reference), **kwargs)

        # Write processing services, if any defined
        self._hsp_write_object_list(h5group, "SERVICES", self._proc_services, "service_", **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a Snapshot object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        sn: ``Snapshot``
            Read Snapshot instance
        """
        # Handle different versions here

        # Read Hdf5StudyPersistent object
        sn = super(Snapshot, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read snapshot time
        t_value = cls._hsp_read_attribute(h5group, 'time', "snapshot time")
        t_unit = cls._hsp_read_unit(h5group, 'time_unit')
        if t_unit is U.none:
            sn.time = t_value
        else:
            sn.time = (t_value, t_unit)

        # Read snapshot physical size
        s_value = cls._hsp_read_attribute(h5group, 'phys_size', "snapshot size")
        s_unit = cls._hsp_read_unit(h5group, 'phys_size_unit')
        if s_unit is U.none:
            sn.physical_size = s_value
        else:
            sn.physical_size = (s_value, s_unit)

        # Read snapshot data reference
        sn.data_reference = cls._hsp_read_attribute(h5group, "data_reference", "data reference")

        # Build data processing service list and add each data processing service into the snapshot, if any defined
        if version >= 2 and "SERVICES" in h5group:
            for service in DataProcessingService._hsp_read_object_list(h5group, "SERVICES", "service_",
                                                                       "snapshot data processing service",
                                                                       dependency_objdict=dependency_objdict):
                sn.processing_services.add(service)

        return sn

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        super(Snapshot, self).galactica_validity_check(**kwargs)

        # Check time value
        if self._time < -10000.0 or self._time > 10000.0:
            log.warning("{sn!s} time value must be defined in the range [-10000.0, 10000.0].".format(sn=self))

        # Check physical size value
        if self._phys_size < 0.0 or self._phys_size > 10000.0:
            log.warning("{sn!s} physical size value must be defined in the range [0.0, 10000.0].".format(sn=self))

        # Check data reference is not too long
        if len(self._data_reference) > 64:
            log.warning("{sn!s} data reference is too long (max. 64 characters).".format(sn=self))

        # Check data reference is filled if linked to a processing service
        if len(self._proc_services) > 0 and len (self._data_reference) == 0:
            log.warning("{sn!s} data reference is not defined while data processing services are bound "
                        "to it.".format(sn=self))

        # Perform validity checks on processing service list
        self._proc_services.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        s = "'{sn_name:s}' snapshot".format(sn_name=self._name)
        return s


__all__ = ["Snapshot"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
Algorithm type
^^^^^^^^^^^^^^

.. autoclass:: astrophysix.simdm.protocol.AlgoType
   :members:
   :undoc-members:

Algorithm
^^^^^^^^^

.. autoclass:: astrophysix.simdm.protocol.Algorithm
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX
"""
from __future__ import absolute_import, unicode_literals
from future.builtins import str, list, dict
import logging
from enum import Enum

from astrophysix.simdm.utils import GalacticaValidityCheckMixin
from astrophysix.utils.strings import Stringifiable
from astrophysix.utils.persistency import Hdf5StudyPersistent

log = logging.getLogger("astrophysix.simdm")


class AlgoType(Enum):
    """
    Algorithm type enum

    Example
    -------
        >>> t = AlgoType.PoissonMultigrid
        >>> t.name
        "Multigrid Poisson solver"
    """
    StructuredGrid = ("struct_grid", "Structured grid method")
    AdaptiveMeshRefinement = ("AMR", "Adaptive mesh refinement")
    SmoothParticleHydrodynamics = ("SPH", "Smooth particle hydrodynamics")
    SpectralMethod = ("spectr_meth", "Spectral method")
    VoronoiMovingMesh = ("Voronoi_MM", "Voronoi tesselation-based moving mesh")
    Godunov = ("Godunov", "Godunov scheme")
    PoissonMultigrid = ("Poisson_MG", "Multigrid Poisson solver")
    PoissonConjugateGradient = ("Poisson_CG", "Conjugate Gradient Poisson solver")
    ParticleMesh = ("PM", "Particle-mesh solver")
    NBody = ("nbody", "N-body method")
    FriendOfFriend = ("FOF", "Friend-of-friend")
    HLLCRiemann = ("HLLC",  "Harten-Lax-van Leer-Contact Riemann solver")
    RayTracer = ("ray_tracer", "Ray-tracer")
    RadiativeTransfer = ("rad_transfer", "Radiative transfer")
    RungeKutta = ("runge_kutta", "Runge-Kutta method")

    def __init__(self, key, name):
        self._key = key
        self._name = name

    @property
    def key(self):
        """Algorithm type indexing key"""
        return self._key

    @property
    def name(self):
        """Algorithm type verbose name"""
        return self._name

    @classmethod
    def from_key(cls, key):
        """
        Parameters
        ----------
        key: :obj:`string`
            algorithm type key

        Returns
        -------
        t: :class:`~astrophysix.simdm.protocol.AlgoType`
            Algorithm type matching the requested key.

        Raises
        ------
        ValueError
            if requested key does not match any algorithm type.

        Example
        -------
            >>> t = AlgoType.from_key("FOF")
            >>> t.name
            "Friend-of-friend"
            >>> t2 = AlgoType.from_key("MY_UNKNOWN_ALGO_TYPE")
            ValuerError: No AlgoType defined with the key 'MY_UNKNOWN_ALGO_TYPE'.
        """
        for algotype in cls:
            if algotype.key == key:
                return algotype
        raise ValueError("No AlgoType defined with the key '{k:s}'.".format(k=key))


class Algorithm(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Protocol algorithm

    Parameters
    ----------
    algo_type: :class:`~astrophysix.simdm.protocol.AlgoType` or :obj:`string`
        AlgoType enum value or AlgoType valid key (mandatory).
    description: :obj:`string`
        algorithm description
    """
    def __init__(self, **kwargs):
        super(Algorithm, self).__init__(**kwargs)

        self._algo_type = AlgoType.AdaptiveMeshRefinement
        self._description = ""

        if "algo_type" not in kwargs:
            raise AttributeError("Algorithm 'algo_type' attribute is not defined (mandatory).")
        atype = kwargs["algo_type"]
        try:
            satype = Stringifiable.cast_string(atype)
            self._algo_type = AlgoType.from_key(satype)
        except ValueError as ve:
            log.error(str(ve))
            raise AttributeError(str(ve))
        except TypeError:  # Not a valid string
            if not isinstance(atype, AlgoType):
                err_msg = "Algorithm 'algo_type' attribute is not a valid AlgoType enum value."
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._algo_type = atype

        if "description" in kwargs:
            self.description = kwargs["description"]

    def __eq__(self, other):
        """
        Algorithm comparison method

        other: :class:`~astrophysix.simdm.protocol.Algorithm`
            algorithm to compare to
        """
        if not super(Algorithm, self).__eq__(other):
            return False

        if self._algo_type != other.algo_type:
            return False

        if self._description != other.description:
            return False

        return True

    @property
    def algo_type(self):
        """Algorithm type (:class:`~astrophysix.simdm.protocol.AlgoType`)"""
        return self._algo_type

    @property
    def name(self):
        """Algorithm name"""
        return self._algo_type.name

    @property
    def description(self):
        """Algorithm description"""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:  # Not a valid string
            raise AttributeError("Algorithm 'description' property is not a valid string.")

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["algo_type", "description"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize an Algorithm object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the Algorithm into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(Algorithm, self)._hsp_write(h5group, **kwargs)

        # Write algorithm type key
        self._hsp_write_attribute(h5group, ('algo_type', self._algo_type.key), **kwargs)

        # Write algorithm description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read an Algorithm object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        algo: ``Algorithm``
            Read Algorithm instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(Algorithm, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read algorithm type
        algotype_key = cls._hsp_read_attribute(h5group, 'algo_type', "algorithm type")

        # Create algorithm object
        algo = cls(uid=uid, algo_type=algotype_key)

        # Read algorithm description, if defined
        algo_descr = cls._hsp_read_attribute(h5group, 'description', "algorithm description",
                                             raise_error_if_not_found=False)
        if algo_descr is not None:
            algo.description = algo_descr

        return algo

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{algo_typename:s}' algorithm".format(algo_typename=self._algo_type.name)


__all__ = ["Algorithm", "AlgoType"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, print_function, division,\
    absolute_import  # Python 2 and 3 compatibility
import logging
import sys


# ------------------------- Logger setup ------------------------------ #
_log = logging.getLogger("astrophysix")

# Info. console message handler
_info_formatter = logging.Formatter(fmt="[%(levelname)s] %(message)s")
_console_handler = logging.StreamHandler(stream=sys.stdout)
_console_handler.setLevel(logging.INFO)
_console_handler.setFormatter(_info_formatter)

# Configure 'astrophysix' logger
_log.setLevel(logging.DEBUG)
_log.addHandler(_console_handler)
# --------------------------------------------------------------------- #


__version__ = "0.7.0"  # Also change in setup.py and doc/source/conf.py
__author__ = 'Damien CHAPON'
__credits__ = 'C.E.A. Saclay - IRFU'



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility

import filecmp

from future.builtins import str, dict
import pytest
import numpy as N
import logging
import os

from astrophysix.simdm import SimulationStudy, ProjectCategory, Project
from astrophysix.simdm.datafiles import Datafile, PlotType, PlotInfo, image, file
from astrophysix.simdm.experiment import Simulation
from astrophysix.simdm.protocol import SimulationCode
from astrophysix.simdm.results import GenericResult
from astrophysix.utils import DatetimeUtil
from astrophysix import units as U
from astrophysix.utils.file import FileType, FileUtil


class TestFileType(object):
    def test_from_key_exceptions(self):
        """
        Tests exceptions raised by FileType.from_alias(method)
        """
        # Check unknown file type
        with pytest.raises(ValueError, match="No FileType defined with the alias 'PLOUC'."):
            ft = FileType.from_alias("PLOUC")

    def test_init(self):
        """
        Tests valid FileType initialisation from aliases
        """
        assert FileType.from_alias("HDF5") == FileType.HDF5_FILE
        assert FileType.from_alias("PNG") == FileType.PNG_FILE
        assert FileType.from_alias("JPEG") == FileType.JPEG_FILE
        assert FileType.from_alias("FITS") == FileType.FITS_FILE
        assert FileType.from_alias("TARGZ") == FileType.TARGZ_FILE
        assert FileType.from_alias("PICKLE") == FileType.PICKLE_FILE
        assert FileType.from_alias("JSON") == FileType.JSON_FILE
        assert FileType.from_alias("CSV") == FileType.CSV_FILE
        assert FileType.from_alias("ASCII") == FileType.ASCII_FILE
        assert FileType.from_alias("YAML") == FileType.YAML_FILE
        assert FileType.from_alias("XML") == FileType.XML_FILE

        # Default HDF5 file extension
        assert FileType.HDF5_FILE.default_extension == ".h5"

        # Pickle file extension regular expression
        assert FileType.PICKLE_FILE.file_regexp.pattern == "(?P<basename>.+)\\.(?P<extension>(pkl|PKL|pickle|sav|save))$"

        # Test valid extension
        assert '.JSON' in FileType.JSON_FILE.extension_list
        assert ".YAML" in FileType.YAML_FILE.extension_list


class TestDatafile(object):
    def test_datafile_init(self):
        """
        Tests Datafile instance initialisation
        """
        # Datafile initialisation
        df = Datafile(name="Gas surface density", description="Gas surface density map of the galactic disk (face on)")
        df[FileType.PNG_FILE] = os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png")

        assert df.name == "Gas surface density"
        assert df.description == "Gas surface density map of the galactic disk (face on)"
        assert df[FileType.PNG_FILE].file_md5sum == "daa130ec42abe0851f78dc73e27900bc"
        assert df[FileType.PNG_FILE].filename == "CEA.png"

        # Test str() conversion
        assert str(df) == "[Gas surface density] datafile"

    def test_setting_datafile_name(self):
        """
        Tests that a datafile defined without name raises an exception
        """
        # Tests datafile defined without name
        with pytest.raises(AttributeError) as e:
            datafile_exc = Datafile()  # No simulation code name defined
        assert str(e.value) == "Datafile 'name' attribute is not defined (mandatory)."

        df = Datafile(name="My plot")

        # ------------- Tests invalid datafile name setting exception ---------------------- #
        empty_protocol_name_err = "Datafile 'name' property is not a valid (non-empty) string."
        with pytest.raises(AttributeError) as e:
            df.name = [-1, 2, 10.5, "Galaxy"]
        assert str(e.value) == empty_protocol_name_err

        with pytest.raises(AttributeError) as e:
            df.name = ""
        assert str(e.value) == empty_protocol_name_err
        # ----------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        df.name = "The best plot ever published"

    def test_setting_datafile_description(self):
        """
        Tests setting datafile description property
        """
        datafile = Datafile(name="Favorite plot", description="Description of my datafile")
        #
        # ------------------------ Tests invalid datafile description setting exception ------------------------------ #
        with pytest.raises(AttributeError) as e:
            datafile.description = 0
        assert str(e.value) == "Datafile 'description' property is not a valid string."
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        datafile.description = "Description of the best datafile in the world"

    def test_setting_datafile_plot_info(self):
        """
        Tests setting datafile plot_info property
        """
        df = Datafile(name="Favorite plot")

        # ------------------------ Tests invalid datafile description setting exception ------------------------------ #
        with pytest.raises(AttributeError, match="Datafile 'plot_info' property is not a valid PlotInfo object."):
            df.plot_info = {"Garbage": 102.5}
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        df.plot_info = PlotInfo(plot_type=PlotType.LINE_PLOT, xaxis_values=N.array([10.0, 20.0, 30.0]),
                                yaxis_values=N.array([1.2, 35.2, 14.9]))

    def test_setting_associated_file(self):
        """
        Add associated files into datafile and tries to get/delete them
        """
        df = Datafile(name="My best datafile")

        # Tests associated file getter and setter
        with pytest.raises(KeyError, match="Datafile 'Salut' key is not a valid FileType index."):
            df["Salut"] = 3.1459

        # Not a valid AssociatedFile object or file path
        with pytest.raises(ValueError,
                           match="Only file paths or PngImageFile objects can be set in Datafile as PNG files."):
            df[FileType.PNG_FILE] = 3.14159

        # Set JPEG file path as PNG file => type mismatch
        jpeg_filepath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "irfu_simple.jpg")
        with pytest.raises(ValueError, match="Datafile associated file type mismatch : expected PngImageFile object "
                                             "but JpegImageFile was provided."):
            df[FileType.PNG_FILE] = image.JpegImageFile.load_file(jpeg_filepath)
        with pytest.raises(AttributeError, match="Invalid filename for a PNG file "
                                                 "\\({fp:s}\\).".format(fp=jpeg_filepath)):
            df[FileType.PNG_FILE] = jpeg_filepath

        # Actually works now it has been implemented...
        # with pytest.raises(NotImplementedError, match="Cannot attach PICKLE file to datafile !"):
        #     df[FileType.PICKLE_FILE] = "/path/to/pickle/file"

        # Valid calls to setter
        df[FileType.PNG_FILE] = os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png")
        df[FileType.JPEG_FILE] = jpeg_filepath
        df[FileType.FITS_FILE] = os.path.join(os.path.dirname(__file__), "io", "datafiles",
                                              "cassiopea_A_0.5-1.5keV.fits")
        df[FileType.TARGZ_FILE] = os.path.join(os.path.dirname(__file__), "io", "datafiles", "archive.tar.gz")
        df[FileType.JSON_FILE] = file.JsonFile.load_file(os.path.join(os.path.dirname(__file__), "io", "datafiles",
                                                                      "test_header_249.json"))
        df[FileType.ASCII_FILE] = os.path.join(os.path.dirname(__file__), "io", "datafiles", "abstract.txt")
        df[FileType.HDF5_FILE] = os.path.join(os.path.dirname(__file__), "io", "python_3", "study.h5")
        df[FileType.YAML_FILE] = os.path.join(os.path.dirname(__file__), "io", "datafiles", "config.yml")
        df[FileType.PICKLE_FILE] = os.path.join(os.path.dirname(__file__), "io", "datafiles", "dict_saved.pkl")

        # Invalid FileType in delitem
        with pytest.raises(KeyError, match="Datafile 'key_1' key is not a valid FileType index."):
            del df["key_1"]

        # Remove PNG image file => should be ok
        del df[FileType.PNG_FILE]
        # Try to remove a PNG file once again, already done that...
        with pytest.raises(KeyError, match="No PNG file is associated to the \\[My best datafile\\] datafile."):
            del df[FileType.PNG_FILE]

        # Invalid FileType in getter
        with pytest.raises(KeyError, match="Datafile 'key_1' key is not a valid FileType index."):
            f = df["key_1"]

        # FileType key not found in getter
        with pytest.raises(KeyError, match="No PNG file is associated to the \\[My best datafile\\] datafile."):
            f = df[FileType.PNG_FILE]

        jpeg = df[FileType.JPEG_FILE]
        assert jpeg.filename == "irfu_simple.jpg" and isinstance(jpeg, image.JpegImageFile)
        json_file = df[FileType.JSON_FILE]
        assert json_file.filename == "test_header_249.json" and isinstance(json_file, file.JsonFile)

    def test_datafile_associated_file_export(self, tmp_path, caplog):
        """
        Tests exporting associated file back into external file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        caplog: captured log PyTest fixture
        """
        df = Datafile(name="My best datafile")

        # Add pickle file and export it, check modification time and perform diff comparison
        orig_pkl_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "dict_saved.pkl")
        df[FileType.PICKLE_FILE] = orig_pkl_fpath
        # Valid but should log an info message
        tmp_fpath = str(tmp_path / df[FileType.PICKLE_FILE].filename)
        fpath = df[FileType.PICKLE_FILE].save_to_disk(tmp_fpath)
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.INFO,
                                            "File '{tfp:s}' saved.".format(tfp=tmp_fpath))
        assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath)) == \
               df[FileType.PICKLE_FILE].last_modified
        assert filecmp.cmp(fpath, orig_pkl_fpath, shallow=False)

        # Add tar.gz file and export it, check modification time and perform diff comparison
        orig_tar_gz_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "archive.tar.gz")
        df[FileType.TARGZ_FILE] = orig_tar_gz_fpath
        tmp_fpath = str(tmp_path / df[FileType.TARGZ_FILE].filename)
        fpath = df[FileType.TARGZ_FILE].save_to_disk(tmp_fpath)
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.INFO,
                                            "File '{tfp:s}' saved.".format(tfp=tmp_fpath))
        assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath)) == \
               df[FileType.TARGZ_FILE].last_modified
        assert filecmp.cmp(fpath, orig_tar_gz_fpath, shallow=False)

        # Add png file and export it, check modification time and perform diff comparison
        orig_png_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png")
        df[FileType.PNG_FILE] = orig_png_fpath
        tmp_fpath = str(tmp_path / df[FileType.PNG_FILE].filename)
        fpath = df[FileType.PNG_FILE].save_to_disk(tmp_fpath)
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.INFO,
                                            "File '{tfp:s}' saved.".format(tfp=tmp_fpath))
        assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath)) == \
               df[FileType.PNG_FILE].last_modified
        assert filecmp.cmp(fpath, orig_png_fpath, shallow=False)

    def test_datafile_equality(self):
        """
        Tests Datafile rich-comparison method Datafile.__eq__()
        """
        df = Datafile(name="Most important datafile",
                      description="This is the best plot ever in the history of Science")

        # Different UUID => not equals
        assert df != Datafile(name=df.name, description=df.description)

        # Different name => not equals
        assert df != Datafile(name="Other name", description=df.description, uid=df.uid)

        # Different description => not equals
        assert df != Datafile(name=df.name, description="Other description", uid=df.uid)

        # Identical datafiles
        df2 = Datafile(name=df.name, description=df.description, uid=df.uid)
        assert df2 == df

        # Plot info comparison
        df.plot_info = PlotInfo(plot_type=PlotType.LINE_PLOT, xaxis_values=N.array([10.0, 20.0, 30.0]),
                                yaxis_values=N.array([1.2, 35.2, 14.9]))
        df2.plot_info = PlotInfo(plot_type=PlotType.HISTOGRAM, xaxis_values=N.array([5.0, 15.0, 25.0, 35.0]),
                                 yaxis_values=N.array([1.25, 5.12, 4.79]))
        assert df != df2
        df.plot_info = df2.plot_info
        assert df == df2

        # Add associated files
        jpeg_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "irfu_simple.jpg")
        png_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png")
        df[FileType.PNG_FILE] = image.PngImageFile.load_file(png_fpath)
        df2[FileType.JPEG_FILE] = jpeg_fpath
        assert df != df2
        df[FileType.JPEG_FILE] = image.JpegImageFile.load_file(jpeg_fpath)
        assert df != df2
        df2[FileType.PNG_FILE] = png_fpath
        assert df == df2

    def test_datafile_galactica_validity_checks(self, caplog):
        """
        Tests datafile Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        df = Datafile(name="Most important datafile")

        # No name defined => should never happen
        # [...]

        # Datafile name too long
        df.name = "This is an overkill long name for the most important datafile ever. xxxxxxxxxxxxxxxxxxxxxxxxxxxx " \
                  "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy"
        df.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[This is an overkill long name for the most important datafile ever. xx"
                                            "xxxxxxxxxxxxxxxxxxxxxxxxxx yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy] "
                                            "datafile name is too long for Galactica (max. 128 characters).")

    def test_datafile_hdf5_io(self, tmp_path):
        """
        Tests saving/loading Datafile from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        # Dummy project
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        res = GenericResult(name="Key result !")
        simu.generic_results.add(res)
        df = res.datafiles.add(Datafile(name="My datafile"))

        # Add tar.gz file
        targz_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "archive.tar.gz")
        df[FileType.TARGZ_FILE] = targz_fpath
        # Add png file
        png_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png")
        df[FileType.PNG_FILE] = png_fpath

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare datafiles
        res_loaded = simu_loaded.generic_results[res.name]
        df_loaded = res_loaded.datafiles[df.name]
        assert df == df_loaded

        # Export associated files and perform diff comparison
        tmp_fpath_targz = str(tmp_path / df_loaded[FileType.TARGZ_FILE].filename)
        fpath = df_loaded[FileType.TARGZ_FILE].save_to_disk(tmp_fpath_targz)
        assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath)) == \
               df[FileType.TARGZ_FILE].last_modified
        assert filecmp.cmp(fpath, targz_fpath, shallow=False)
        tmp_fpath_png = str(tmp_path / df_loaded[FileType.PNG_FILE].filename)
        fpath = df_loaded[FileType.PNG_FILE].save_to_disk(tmp_fpath_png)
        assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath)) == \
               df[FileType.PNG_FILE].last_modified
        assert filecmp.cmp(fpath, png_fpath, shallow=False)

        # ------------------------------------------------------------------------------------------------------------ #
        # Try to add a new JPEG associated file in the loaded study and saves it again
        jpeg_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "irfu_simple.jpg")
        df_loaded[FileType.JPEG_FILE] = jpeg_fpath
        fname2 = str(tmp_path / "study_saved_again.h5")
        study_loaded.save_HDF5(fname2)

        # Now reloads the study once again and compare datafiles
        study_reloaded = SimulationStudy.load_HDF5(fname2)
        df_reloaded = study_reloaded.project.simulations[simu.name].generic_results[res.name].datafiles[df.name]
        assert df_loaded == df_reloaded

        # Try to edit the PNG associated file in the loaded study and saves it again
        png_fpath_black = os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA_black.png")  # Different CEA logo with black letters
        df_loaded[FileType.PNG_FILE] = png_fpath_black
        study_loaded.save_HDF5(fname2)

        # Now reloads the study once again and compare datafiles
        study_reloaded = SimulationStudy.load_HDF5(fname2)

        df_reloaded = study_reloaded.project.simulations[simu.name].generic_results[res.name].datafiles[df.name]
        assert df_loaded == df_reloaded

        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Finally, compare exported files with the original one ~~~~~~~~~~~~~~~~~~~~~~~~ #
        # --> First TARGZ file saved in the first ever saved study
        fpath_targz = df_reloaded[FileType.TARGZ_FILE].save_to_disk(str(tmp_path / df_reloaded[FileType.TARGZ_FILE].filename))
        assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath_targz)) == \
               df[FileType.TARGZ_FILE].last_modified
        assert filecmp.cmp(fpath_targz, targz_fpath, shallow=False)

        # --> Second, the PNG file edited on the study loaded once
        fpath_png = df_reloaded[FileType.PNG_FILE].save_to_disk(str(tmp_path / df_reloaded[FileType.PNG_FILE].filename))
        # assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath_png)) != \
        #        df[FileType.PNG_FILE].last_modified
        assert not filecmp.cmp(fpath_png, png_fpath, shallow=False)
        assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath_png)) == \
               df_loaded[FileType.PNG_FILE].last_modified
        assert filecmp.cmp(fpath_png, png_fpath_black, shallow=False)

        # --> Third, the JPEG file saved on the study loaded once
        fpath_jpg = df_reloaded[FileType.JPEG_FILE].save_to_disk(str(tmp_path / df_reloaded[FileType.JPEG_FILE].filename))
        # assert DatetimeUtil.utc_from_timestamp(FileUtil.last_modification_timestamp(fpath_jpg)) == \
        #        df_loaded[FileType.JPEG_FILE].last_modified
        assert filecmp.cmp(fpath_jpg, jpeg_fpath, shallow=False)
        # ------------------------------------------------------------------------------------------------------------ #

        new_study = SimulationStudy(project=study_reloaded.project)
        new_study.save_HDF5(str(tmp_path / "test_rewrite.h5"))


class TestPlotType(object):
    def test_from_key_exceptions(self):
        """
        Tests exceptions raised by PlotType.from_alias(method)
        """
        # Check unknown plot type
        with pytest.raises(ValueError, match="No PlotType defined with the alias 'PLOUC'."):
            pt = PlotType.from_alias("PLOUC")

    def test_init(self):
        """
        Tests valid PlotType initialisation from aliases
        """
        assert PlotType.from_alias("line") == PlotType.LINE_PLOT
        assert PlotType.from_alias("hist") == PlotType.HISTOGRAM
        assert PlotType.from_alias("2d_hist") == PlotType.HISTOGRAM_2D
        assert PlotType.from_alias("scatter") == PlotType.SCATTER_PLOT
        assert PlotType.from_alias("img") == PlotType.IMAGE
        assert PlotType.from_alias("2d_map") == PlotType.MAP_2D

        # PlotType display name
        assert PlotType.HISTOGRAM_2D.display_name == "2D histogram"

        # PlotType number of dimensions
        assert PlotType.HISTOGRAM_2D.ndimensions == 2
        assert PlotType.LINE_PLOT.ndimensions == 1
        assert PlotType.LINE_PLOT.axis_size_offset == 0
        assert PlotType.MAP_2D.axis_size_offset == 1


class TestPlotInfo(object):
    def test_plot_info_init(self):
        """
        Tests PlotInfo object initialisation
        """
        # 1D histogram
        pinfo = PlotInfo(plot_type=PlotType.HISTOGRAM, yaxis_values=N.array([1.0e3, 2.0e4]),
                         xaxis_values=N.array([1.2, 2.4, 3.6]), xaxis_log_scale=False, yaxis_log_scale=True,
                         xlabel="my label for x", ylabel="my label for y", title="My plot", xaxis_unit="Myr",
                         yaxis_unit="kpc")

        assert pinfo.title == "My plot"
        assert pinfo.plot_type == PlotType.HISTOGRAM
        assert (pinfo.yaxis_values == N.array([1.0e3, 2.0e4])).all()
        assert (pinfo.xaxis_values == N.array([1.2, 2.4, 3.6])).all()
        assert pinfo.xlabel == "my label for x"
        assert pinfo.ylabel == "my label for y"
        assert not pinfo.xaxis_log_scale
        assert pinfo.yaxis_log_scale
        assert pinfo.xaxis_unit == U.Myr
        assert pinfo.yaxis_unit == U.kpc

        # 2D histogram
        pinfo_2d = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, yaxis_values=N.array([10.0, 20.0, 30.0]), title="My plot2",
                            xaxis_values=N.array([1.0, 2.0, 3.0]), values=N.array([[1.2e3, 5.5e4], [2.4e4, 9.1e3]]),
                            xaxis_log_scale=False, yaxis_log_scale=False, values_log_scale=True,
                            xlabel="my label for x", ylabel="my label for y", values_label="total mass",
                            xaxis_unit="kpc", yaxis_unit="pc", values_unit=U.Msun)

        assert pinfo_2d.title == "My plot2"
        assert pinfo_2d.plot_type == PlotType.HISTOGRAM_2D
        assert (pinfo_2d.xaxis_values == N.array([1.0, 2.0, 3.0])).all()
        assert (pinfo_2d.yaxis_values == N.array([10.0, 20.0, 30.0])).all()
        assert (pinfo_2d.values == N.array([[1.2e3, 5.5e4], [2.4e4, 9.1e3]])).all()
        assert pinfo_2d.xlabel == "my label for x"
        assert pinfo_2d.ylabel == "my label for y"
        assert pinfo_2d.values_label == "total mass"
        assert not pinfo.xaxis_log_scale
        assert not pinfo_2d.yaxis_log_scale
        assert pinfo_2d.values_log_scale
        assert pinfo_2d.xaxis_unit == U.kpc
        assert pinfo_2d.yaxis_unit == U.pc
        assert pinfo_2d.values_unit == U.Msun

    def test_setting_plot_info_type(self):
        """
        Tests PlotInfo object creation with valid /invalid 'plot_type' attribute
        """
        # PlotInfo defined without 'plot_type' mandatory attribute => Exception
        with pytest.raises(AttributeError, match="PlotInfo 'plot_type' attribute is not defined \\(mandatory\\)."):
            pi = PlotInfo()

        # Tests invalid plot_type property initialisation exception
        with pytest.raises(AttributeError, match="No PlotType defined with the alias 'TRICOT'."):
            PlotInfo(plot_type="TRICOT")  # Unknown plot type

        with pytest.raises(AttributeError, match="PlotInfo 'plot_type' attribute is not a valid PlotType enum value."):
            PlotInfo(plot_type=(3, "Strawberries"))  # Invalid plot_type

    def test_plot_info_data(self):
        """
        Tests PlotInfo instance data setup with PlotInfo.set_data() method
        """
        with pytest.raises(AttributeError, match="PlotInfo 'xaxis_values' attribute is not defined \\(mandatory\\)."):
            pi = PlotInfo(plot_type=PlotType.LINE_PLOT)
        with pytest.raises(AttributeError, match="PlotInfo 'yaxis_values' attribute is not defined \\(mandatory\\)."):
            pi = PlotInfo(plot_type=PlotType.LINE_PLOT, xaxis_values=N.array([1., 2., 3., 4.]))

        # Invalid yaxis_values attribute
        values_1d_arr_error = "'yaxis_values' PlotInfo attribute must be a 1-dimensional array."
        with pytest.raises(AttributeError, match=values_1d_arr_error):
            pi = PlotInfo(plot_type=PlotType.LINE_PLOT, yaxis_values=N.array([[1., 2.], [3., 4.]]),  # 2D array here !
                          xaxis_values=N.array([10., 20., 30., 40.]))
        with pytest.raises(AttributeError, match=values_1d_arr_error):
            pi = PlotInfo(plot_type=PlotType.LINE_PLOT, yaxis_values={"a": [1, 2, 3, 4]},  # A dict here !
                          xaxis_values=N.array([10., 20., 30., 40.]))

        # Invalid xaxis_values attribute
        xaxis_values_1d_arr_error = "'xaxis_values' PlotInfo attribute must be a 1-dimensional array."
        with pytest.raises(AttributeError, match=xaxis_values_1d_arr_error):
            pi = PlotInfo(plot_type=PlotType.LINE_PLOT, yaxis_values=N.array([1., 2., 3., 4.]),
                          xaxis_values={3.5: [2.3, 4.5]})  # Adict here !
        with pytest.raises(AttributeError, match=xaxis_values_1d_arr_error):
            pi = PlotInfo(plot_type=PlotType.LINE_PLOT, yaxis_values=N.array([1., 2., 3., 4.]),
                          xaxis_values=N.array([[10., 20.], [30., 40.]]))  # 2D array here !

        # Array size mismatch (1D plots)
        arr_size_mismatch_err = "Array size mismatch : 'yaxis_values' coordinate array size should be 5 \\(x-axis " \
                                "coordinate array size=5\\) for 'Line plot'."
        with pytest.raises(AttributeError, match=arr_size_mismatch_err):
            pi = PlotInfo(plot_type=PlotType.LINE_PLOT, yaxis_values=N.array([1., 2., 3., 4.]),
                          xaxis_values=N.array([10., 20., 30., 40., 50.]))
        arr_size_mismatch_err = "Array size mismatch : 'yaxis_values' coordinate array size should be 3 \\(x-axis " \
                                "coordinate array size=4\\) for 'Histogram'."
        with pytest.raises(AttributeError, match=arr_size_mismatch_err):
            pi = PlotInfo(plot_type=PlotType.HISTOGRAM, yaxis_values=N.array([1., 2., 3., 4.]),
                          xaxis_values=N.array([10., 20., 30., 40.]))

        # Invalid 'values' attribute (2D plots)
        with pytest.raises(AttributeError, match="PlotInfo 'values' attribute is not defined \\(mandatory for 2D "
                                                 "plots\\)."):
            pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, xaxis_values=N.array([10., 20., 30.]),
                          yaxis_values=N.array([10., 20., 30.]))  # Missing values attribute
        values_1d_arr_error = "PlotInfo 'values' attribute must be a 2-dimensional array."
        with pytest.raises(AttributeError, match=values_1d_arr_error):
            pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, xaxis_values=N.array([10., 20., 30.]),
                          yaxis_values=N.array([10., 20., 30.]), values={1: 'a'})  # A dict here !
        with pytest.raises(AttributeError, match=values_1d_arr_error):
            pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, xaxis_values=N.array([10., 20., 30.]),
                          yaxis_values=N.array([10., 20., 30.]), values=N.array([10.0, 20.0, 30.0, 40.0]))  # 1D array here !

        # Array size mismatch
        arr_size_mismatch_err = "Array size mismatch : 'values' array \\(shape=\\(2, 2\\)\\) should have a shape " \
                                "\\(2, 3\\) \\(x-axis coordinate array size=3 ; y-axis coordinate array size=4\\) " \
                                "for '2D histogram'."
        with pytest.raises(AttributeError, match=arr_size_mismatch_err):
            pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, values=N.array([[1., 2.], [3., 4.]]),
                          xaxis_values=N.array([10., 20., 30.]), yaxis_values=N.array([10.0, 20.0, 30.0, 40.0]))

        # Should be ok
        pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, values=N.array([[1., 2.], [3., 4.]]),
                      xaxis_values=N.array([10., 20., 30.]), yaxis_values=N.array([10.0, 20.0, 30.0]))
        pi.set_data(N.array([1.0, 2.0, 3.0, 4.0]), N.array([1.0, 2.0, 3.0, 4.0]),
                    values=N.array([[1.5, 6.2, 32.1], [6.4, 8.1, 0.6], [4.7, 22.1, 15.4]]))

    def test_plot_info_unavailable_attrs_1d_plots(self):
        """
        Tests unavailable PlotInfo attributes for 1D plots (only available for 2D plots)
        """
        pinfo = PlotInfo(plot_type=PlotType.SCATTER_PLOT, yaxis_values=N.array([1.0e3, 2.0e4, 4.5e5]),
                         xaxis_values=N.array([1.2, 2.4, 3.6]))

        with pytest.raises(AttributeError, match="PlotInfo object does not have a 'values' property."):
            v = pinfo.values

        # Values log-scale flag  property (gettter + setter)
        err_log_scale = "PlotInfo object does not have a 'values_log_scale' property."
        with pytest.raises(AttributeError, match=err_log_scale):
            vls = pinfo.values_log_scale
        with pytest.raises(AttributeError, match=err_log_scale):
            pinfo.values_log_scale = True

        # Values unit  property (gettter + setter)
        err_values_unit = "PlotInfo object does not have a 'values_unit' property."
        with pytest.raises(AttributeError, match=err_values_unit):
            vu = pinfo.values_unit
        with pytest.raises(AttributeError, match=err_values_unit):
            pinfo.values_unit = U.Msun

        err_values_label = "PlotInfo object does not have a 'values_label' property."
        with pytest.raises(AttributeError, match=err_values_label):
            vl = pinfo.values_label
        with pytest.raises(AttributeError, match=err_values_label):
            pinfo.values_label = "my label"

    def test_plot_info_units(self):
        """
        Tests PlotInfo unit setup (x-axis/y-axis/values)
        """
        pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, values=N.array([[1., 2.], [3., 4.]]),
                      xaxis_values=N.array([10., 20., 30.]), yaxis_values=N.array([10.0, 20.0, 30.0]))

        # Invalid x-axis unit values
        inv_unit_err = "PlotInfo 'xaxis_unit' property is not a valid \\(non-empty\\) string."
        with pytest.raises(AttributeError, match=inv_unit_err):
            pi.xaxis_unit = ""
        with pytest.raises(AttributeError, match=inv_unit_err):
            pi.xaxis_unit = 153.25
        with pytest.raises(AttributeError, match="PlotInfo 'xaxis_unit' property error : Unknown unit name 'Smurf'."):
            pi.xaxis_unit = "Smurf"

        # Invalid y-axis unit values
        inv_unit_err = "PlotInfo 'yaxis_unit' property is not a valid \\(non-empty\\) string."
        with pytest.raises(AttributeError, match=inv_unit_err):
            pi.yaxis_unit = ""
        with pytest.raises(AttributeError, match=inv_unit_err):
            pi.yaxis_unit = 153.25
        with pytest.raises(AttributeError, match="PlotInfo 'yaxis_unit' property error : Unknown unit name 'Pluto'."):
            pi.yaxis_unit = "Pluto"

        # Invalid x-axis unit values
        inv_unit_err = "PlotInfo 'values_unit' property is not a valid \\(non-empty\\) string."
        with pytest.raises(AttributeError, match=inv_unit_err):
            pi.values_unit = ""
        with pytest.raises(AttributeError, match=inv_unit_err):
            pi.values_unit = 153.25
        with pytest.raises(AttributeError, match="PlotInfo 'values_unit' property error : Unknown unit name 'Mickey'."):
            pi.values_unit = "Mickey"

        # Ok
        pi.xaxis_unit = U.Myr
        pi.yaxis_unit = U.kpc
        pi.values_unit = U.Msun

    def test_plot_info_title_labels(self):
        """
        Tests PlotInfo label setup (x-axis/y-axis/value) and title setup
        """
        pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, values=N.array([[1., 2.], [3., 4.]]),
                      xaxis_values=N.array([10., 20., 30.]), yaxis_values=N.array([10.0, 20.0, 30.0]))

        # Invalid x-axis label
        inv_label_err = "PlotInfo 'xlabel' property is not a valid string."
        with pytest.raises(AttributeError, match=inv_label_err):
            pi.xlabel = 123.5

        # Invalid y-axis label
        inv_label_err = "PlotInfo 'ylabel' property is not a valid string."
        with pytest.raises(AttributeError, match=inv_label_err):
            pi.ylabel = 123.5

        # Invalid x-axis label
        inv_label_err = "PlotInfo 'values_label' property is not a valid string."
        with pytest.raises(AttributeError, match=inv_label_err):
            pi.values_label = 123.5

        # Invalid plot title
        inv_label_err = "PlotInfo 'title' property is not a valid string."
        with pytest.raises(AttributeError, match=inv_label_err):
            pi.title = 123.5

    def test_plot_info_log_scale(self):
        """
        Tests PlotInfo log scale flags (x-axis/y-axis/value)
        """
        pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, values=N.array([[1., 2.], [3., 4.]]),
                      xaxis_values=N.array([10., 20., 30.]), yaxis_values=N.array([10.0, 20.0, 30.0]))

        with pytest.raises(AttributeError, match="'xaxis_log_scale' PlotInfo property must be a boolean value."):
            pi.xaxis_log_scale = 0.5
        pi.xaxis_log_scale = True
        with pytest.raises(AttributeError, match="'yaxis_log_scale' PlotInfo property must be a boolean value."):
            pi.yaxis_log_scale = 0.5
        pi.yaxis_log_scale = False
        with pytest.raises(AttributeError, match="'values_log_scale' PlotInfo property must be a boolean value."):
            pi.values_log_scale = 0.5
        pi.values_log_scale = True

    def test_plot_info_equality(self):
        """
        Tests PlotInfo instance equality
        """
        # 2D histogram
        pi = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, yaxis_values=N.array([10.0, 20.0, 30.0]),
                      xaxis_values=N.array([1.0, 2.0, 3.0]), values=N.array([[1.2e3, 5.5e4], [2.4e4, 9.1e3]]))

        # Different UUID => not equals
        assert pi != PlotInfo(plot_type=pi.plot_type, yaxis_values=pi.yaxis_values, xaxis_values=pi.xaxis_values,
                              values=pi.values)

        # Different type => not equals
        assert pi != PlotInfo(plot_type=PlotType.LINE_PLOT, yaxis_values=pi.yaxis_values, xaxis_values=pi.xaxis_values,
                              values=pi.values, uid=pi.uid)

        # Different title => not equals
        pi2 = PlotInfo(plot_type=pi.plot_type, yaxis_values=pi.yaxis_values, xaxis_values=pi.xaxis_values,
                       values=pi.values, uid=pi.uid)
        assert pi == pi2
        pi.title = "My plot"
        pi2.title = "Test"
        assert pi != pi2
        pi2.title = pi.title
        assert pi == pi2

        # Different axis/value labels => not equals
        pi.xlabel = "x-axis label"
        pi2.xlabel = "dummy"
        assert pi != pi2
        pi2.xlabel = pi.xlabel
        assert pi == pi2
        pi.ylabel = "y-axis label"
        pi2.ylabel = "dummy"
        assert pi != pi2
        pi2.ylabel = pi.ylabel
        assert pi == pi2
        pi.values_label = "value label"
        pi2.values_label = "dummy"
        assert pi != pi2
        pi2.values_label = pi.values_label
        assert pi == pi2

        # Different axis/value log-scale flags => not equals
        pi.xaxis_log_scale = True
        pi2.xaxis_log_scale = False
        assert pi != pi2
        pi2.xaxis_log_scale = pi.xaxis_log_scale
        assert pi == pi2
        pi.yaxis_log_scale = True
        pi2.yaxis_log_scale = False
        assert pi != pi2
        pi2.yaxis_log_scale = pi.yaxis_log_scale
        assert pi == pi2
        pi.values_log_scale = True
        pi2.values_log_scale = False
        assert pi != pi2
        pi2.values_log_scale = pi.values_log_scale
        assert pi == pi2

        # Unit not equals => not equals
        pi.xaxis_unit = U.kpc
        pi2.xaxis_unit = 1.0E3*U.pc
        assert pi != pi2
        pi2.xaxis_unit = U.kpc
        assert pi == pi2
        pi.yaxis_unit = U.kpc
        pi2.yaxis_unit = 1.0E3 * U.pc
        assert pi != pi2
        pi2.yaxis_unit = U.kpc
        assert pi == pi2
        pi.values_unit = U.Msun
        pi2.values_unit = U.kg
        assert pi != pi2
        pi2.values_unit = U.Msun
        assert pi == pi2

        # Different data and MD5sums
        pi2.set_data(pi.xaxis_values, pi.yaxis_values, pi.values-1.0)
        assert pi != pi2
        pi2.set_data(pi.xaxis_values, pi.yaxis_values-1.0, pi.values)
        assert pi != pi2
        pi2.set_data(pi.xaxis_values-1.0, pi.yaxis_values, pi.values)
        assert pi != pi2
        pi2.set_data(pi.xaxis_values, pi.yaxis_values, pi.values)
        assert pi == pi2

    def test_plot_info_hdf5_io(self, tmp_path):
        """
        Tests saving/loading PlotInfo from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path PyTest fixture
        """
        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = proj.simulations.add(Simulation(simu_code=ramses, name="My simu énorme"))
        res = simu.generic_results.add(GenericResult(name="Key result !"))
        df = res.datafiles.add(Datafile(name="My datafile"))
        df.plot_info = PlotInfo(plot_type=PlotType.HISTOGRAM_2D, yaxis_values=N.array([10.0, 20.0, 30.0]), title="plot",
                                xaxis_values=N.array([1.0, 2.0, 3.0]), values=N.array([[1.2e3, N.nan], [2.4e4, 9.1e3]]),
                                xaxis_log_scale=True, yaxis_log_scale=True, values_log_scale=True,
                                xlabel="my label for x", ylabel="my label for y", values_label="total mass",
                                xaxis_unit=U.kpc, yaxis_unit=U.Mpc, values_unit=U.Msun)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare PlotInfo objects
        res_loaded = simu_loaded.generic_results[res.name]
        df_loaded = res_loaded.datafiles[df.name]
        assert df.plot_info == df_loaded.plot_info


__all__ = ["TestFileType", "TestDatafile", "TestPlotType", "TestPlotInfo"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
Physics
^^^^^^^

.. autoclass:: astrophysix.simdm.protocol.Physics
   :members:
   :undoc-members:


Physical process
^^^^^^^^^^^^^^^^

.. autoclass:: astrophysix.simdm.protocol.PhysicalProcess
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX
"""
from __future__ import absolute_import, unicode_literals
from future.builtins import str, list, dict
import logging
from enum import Enum

from astrophysix.simdm.utils import GalacticaValidityCheckMixin
from astrophysix.utils.strings import Stringifiable
from astrophysix.utils.persistency import Hdf5StudyPersistent

log = logging.getLogger("astrophysix.simdm")


class Physics(Enum):
    """
    Physics enum

    Example
    -------
        >>> ph = Physics.MHD
        >>> ph.name
        "Magnetohydrodynamics"
    """
    SelfGravity = ("self_gravity", "Self-gravity")
    ExternalGravity = ("ext_gravity", "External gravity")
    Hydrodynamics = ("hydro",  "Hydrodynamics")
    MHD = ("mhd", "Magnetohydrodynamics")
    StarFormation = ("star_form", "Star formation")
    SupernovaeFeedback = ("sn_feedback", "Supernovae feedback")
    SupermassiveBlackHoleFeedback = ("smbh_feedback", "SMBH feedback")
    StellarIonisingRadiation = ("stell_ion_rad", "Stellar ionising radiation")
    StellarUltravioletRadiation = ("stell_uv_rad", "Stellar ultraviolet radiation")
    StellarInfraredRadiation = ("stell_ir_rad", "Stellar infrared radiation")
    AGNFeedback = ("AGN_feedback", "AGN feedback")
    ProtostellarJetFeedback = ("psjet_feedback", "Protostellar jet feedback")
    Chemistry = ("chemistry", "Chemistry")
    DustCooling = ("dust_cooling", "Dust cooling")
    MolecularCooling = ("mol_cooling", "Molecular cooling")
    AtomicCooling = ("atomic_cooling", "Atomic cooling")
    TurbulentForcing = ("turb_forcing", "Turbulent forcing")
    RadiativeTransfer = ("rad_transfer", "Radiative transfer")

    def __init__(self, key, name):
        self._key = key
        self._name = name

    @property
    def key(self):
        """Physics indexing key"""
        return self._key

    @property
    def name(self):
        """PHysics verbose name"""
        return self._name

    @classmethod
    def from_key(cls, key):
        """
        Parameters
        ----------
        key: :obj:`string`
            physics key

        Returns
        -------
        t: :class:`~astrophysix.simdm.protocol.Physics`
            Physics matching the requested key.

        Raises
        ------
        ValueError
            if requested key does not match any physics.

        Example
        -------
            >>> ph = Physics.from_key("star_from")
            >>> ph.name
            "Star formation"
            >>> ph2 = Physics.from_key("MY_UNKNOWN_PHYSICS")
            ValuerError: No Physics defined with the key 'MY_UNKNOWN_PHYSICS'.
        """
        for ph in cls:
            if ph.key == key:
                return ph
        raise ValueError("No Physics defined with the key '{k:s}'.".format(k=key))


class PhysicalProcess(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Simulation code physical process

    Parameters
    ----------
    physics: :class:`~astrophysix.simdm.protocol.Physics` or :obj:`string`
        Physics enum value or Physics valid key. (mandatory)
    description: :obj:`string`
        physics description
    """
    def __init__(self, **kwargs):
        super(PhysicalProcess, self).__init__(**kwargs)

        self._physics = Physics.StarFormation
        self._description = ""

        if "physics" not in kwargs:
            raise AttributeError("PhysicalProcess 'physics' attribute is not defined (mandatory).")
        ph = kwargs["physics"]
        try:
            sph = Stringifiable.cast_string(ph)
            self._physics = Physics.from_key(sph)
        except ValueError as ve:
            log.error(str(ve))
            raise AttributeError(str(ve))
        except TypeError:
            if not isinstance(ph, Physics):
                err_msg = "PhysicalProcess 'physics' attribute is not a valid Physics enum value."
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._physics = ph

        if "description" in kwargs:
            self.description = kwargs["description"]

    def __eq__(self, other):
        """
        PhysicalProcess comparison method

        other: :class:`~astrophysics.simdm.protocol.PhysicalProcess`
            physical process to compare to
        """
        if not super(PhysicalProcess, self).__eq__(other):
            return False

        if self._physics != other.physics:
            return False

        if self._description != other.description:
            return False

        return True

    @property
    def physics(self):
        """Prrocess type (:class:`~astrophysix.simdm.protocol.Physics`)"""
        return self._physics

    @property
    def name(self):
        """Physical process name"""
        return self._physics.name

    @property
    def description(self):
        """Physical process description"""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            raise AttributeError("PhysicalProcess 'description' property is not a valid string.")

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["physics", "description"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a PhysicalProcess object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the PhysicalProcess into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(PhysicalProcess, self)._hsp_write(h5group, **kwargs)

        # Write physics type key
        self._hsp_write_attribute(h5group, ('phys_type', self._physics.key), **kwargs)

        # Write physical process description, if  defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a PhysicalProcess object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        exp: ``PhysicalProcess``
            Read PhysicalProcess instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(PhysicalProcess, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read physics type
        phys_key = cls._hsp_read_attribute(h5group, 'phys_type', "physical process type")

        # Create physical process object
        pproc = cls(uid=uid, physics=phys_key)

        # Read physical process description, if defined
        pp_desc = cls._hsp_read_attribute(h5group, 'description', "physical process description",
                                          raise_error_if_not_found=False)
        if pp_desc is not None:
            pproc.description = pp_desc

        return pproc

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{phys:s}' physical process".format(phys=self._physics.name)


__all__ = ["PhysicalProcess", "Physics"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
@startwbs

+ Simulation study
 + Project
  + Target objects
   +_ Spiral galaxy
   +_ Elliptical galaxy
  + Simulation codes
   +_ RAMSES
   +_ AREPO
  + Simulations
   +_ [RAMSES] Simulation #1
    + Generic results
     +_ Generic Result #1
      + Datafiles
       +_ Star formation history plot
     +_ Result #2
    + Snapshots
     +_ Snapshot #234
      + Datafiles
       +_ Gas density 2D map
        +_ 'rho_dens.png' PNG file
        +_ 'rho_slice.fits' FITS file
       +_ Star formation radial 1D profile
      + Catalogs
       +_ [Spiral galaxy] Galaxy catalog (Z=2.0)
     +_ Snapshot #587
  + Post-processing runs
   + Post-processing codes
    +_ RadMC
   +_ [RadMC] Post-pro. run #1 (Simulation #2)
    + Generic results
     +_ Synthetic observed map
      + Datafiles
       +_ Map #1
        +_ 'map.jpeg' JPEG file
   +_ [AREPO] Simulation #2

@endwbs
"""
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility

from enum import Enum

from future.builtins import str, list
import logging

from astrophysix.utils.persistency import Hdf5StudyPersistent
from .datafiles import PngImageFile, JpegImageFile
from .datafiles.image import ImageFile
from .utils import ObjectList, GalacticaValidityCheckMixin
from .catalogs.targobj import TargetObject
from astrophysix.utils.strings import Stringifiable
from .experiment import Simulation
from .protocol import SimulationCode, PostProcessingCode, InputParameter


__doc__ = """

.. autoclass:: astrophysix.simdm.ProjectCategory
   :members:
   :undoc-members:

.. autoclass:: astrophysix.simdm.Project
   :members:
   :undoc-members:

"""

from ..utils import FileUtil, FileType

log = logging.getLogger("astrophysix.simdm")


class ProjectCategory(Enum):
    """
    Project category enum

    Example
    -------
        >>> cat = ProjectCategory.PlanetaryAtmospheres
        >>> cat.verbose_name
        "Planetary atmospheres"
    """
    SolarPhysics = ("SOLAR_PHYSICS", "Solar Physics")
    PlanetaryAtmospheres = ("PLANET_ATMO", "Planetary atmospheres")
    StellarEnvironments = ("STELLAR_ENVS", "Stellar environments")
    StarPlanetInteractions = ("STAR_PLANET_INT", "Star-planet interactions")
    StarFormation = ("STAR_FORM", "Star formation")
    InterstellarMedium = ("ISM", "Interstellar medium")
    Magnetohydrodynamics = ("MHD", "Magnetohydrodynamics")
    Supernovae = ("SUPERNOVAE", "Supernovae")
    GalacticDynamics = ("GALAXIES", "Galactic dynamics")
    Cosmology = ("COSMOLOGY", "Cosmology")

    def __init__(self, alias, verbose):
        self._alias = alias
        self._verbose = verbose

    @property
    def alias(self):
        """Project category alias"""
        return self._alias

    @property
    def verbose_name(self):
        """Project category verbose name"""
        return self._verbose

    @classmethod
    def from_alias(cls, alias):
        """

        Parameters
        ----------
        alias: :obj:`string`
            project category alias

        Returns
        -------
        c: :class:`~astrophysix.simdm.ProjectCategory`
            Project category matching the requested alias.

        Raises
        ------
        ValueError
            if requested alias does not match any project category.

        Example
        -------
            >>> c = ProjectCategory.from_alias("STAR_FORM")
            >>> c.verbose_name
            "Star formation"
            >>> c2 = ProjectCategory.from_alias("MY_UNKNOWN_CATEGORY")
            ValuerError: No ProjectCategory defined with the alias 'MY_UNKNOWN_CATEGORY'.
        """
        for cat in cls:
            if cat.alias == alias:
                return cat
        raise ValueError("No ProjectCategory defined with the alias '{a:s}'.".format(a=alias))


class Project(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Numerical study project (Simulation data model)

    Parameters
    ----------
    category: :class:`~astrophysix.simdm.project.ProjectCategory` or :obj:`string`
        project category or project category alias (mandatory)
    project_title: :obj:`string`
        project title (mandatory)
    alias: :obj:`string`
        Project alias (if defined, 16 max characters is recommended)
    short_description: :obj:`string`
        project short description
    general_description: :obj:`string`
        (long) project description
    data_description: :obj:`string`
        available data description in the project
    acknowledgement: :obj:`string`
        Project acknowledgement text. describes how to acknowledge the work presented in this project in any
        publication.
    directory_path: :obj:`string`
        project directory path
    thumbnail_image: :obj:`string` or :class:`~astrophysix.simdm.datafiles.image.ImageFile`
        project image thumbnail for fancy display on the Galactica simulation database.
        
    Example
    -------

        >>> # Define a new project
        >>> proj = Project(project_title="My M51 galaxy interaction model", category=ProjectCategory.GalaxyMergers,
        ...                alias="M51_SIM2022", short_description="M51 interacting galaxy model at 10 pc scale with" 
        ...                "low-efficiency star formation", general_description=gen_descr, data_description=data_descr,
        ...                directory_path="/raid/data/PROJS/M51_M/", thumbnail_image="~/Pictures/M51_M/thumb_M51.png",
        ...                acknowledgement="Please cite Bournaud et al. 2022 (in prep.) if you use data from this project.")
        >>>
        >>> # Set the project in a study
        >>> study = SimulationStudy(project=proj)
    """
    # _hsp_version = 2  # Project with target objects + acknowledgement property
    # _hsp_version = 3  # Project with thumbnail PNG/JPEG image file (optional)
    _hsp_version = 4  # Project with simulation datatable parameter list
    def __init__(self, *args, **kwargs):
        uid = kwargs.pop("uid", None)
        super(Project, self).__init__(uid=uid, **kwargs)

        self._category = ProjectCategory.StarFormation
        self._title = ""
        self._short_description = ""
        self._general_description = ""
        self._data_description = ""
        self._alias = ""
        self._how_to_acknowledge = ""
        self._directory_path = ""
        self._thumbnail_img = None

        self._simulations = ObjectList(Simulation, "name",
                                       object_addition_delhandler=(self._can_delete_input_param,
                                                                   ["simulation_code", "input_parameters"]))
        self._simu_dt_params = ObjectList(InputParameter, 'name')
        self._simu_dt_params.add_validity_check_method(self._check_valid_simu_input_param)

        if "category" not in kwargs:
            raise AttributeError("Project 'category' attribute is not defined (mandatory).")
        self.category = kwargs["category"]

        if "project_title" not in kwargs:
            raise AttributeError("Project 'project_title' attribute is not defined (mandatory).")
        self.project_title = kwargs["project_title"]

        if "alias" in kwargs:
            self.alias = kwargs["alias"]

        if "short_description" in kwargs:
            self.short_description = kwargs["short_description"]

        if "general_description" in kwargs:
            self.general_description = kwargs["general_description"]

        if "data_description" in kwargs:
            self.data_description = kwargs["data_description"]

        if "acknowledgement" in kwargs:
            self._how_to_acknowledge = kwargs["acknowledgement"]

        if "directory_path" in kwargs:
            self.directory_path = kwargs["directory_path"]

        if "thumbnail_image" in kwargs:
            self.thumbnail_image = kwargs["thumbnail_image"]

    def __eq__(self, other):
        """
        Project comparison method

        other: :class:`~astrophysix.simdm.Project`
            project to compare to.
        """
        if not super(Project, self).__eq__(other):
            return False

        if self._category != other.category:
            return False

        if self._title != other.project_title:
            return False

        if self._alias != other.alias:
            return False

        if self._short_description != other.short_description:
            return False

        if self._general_description != other.general_description:
            return False

        if self._data_description != other.data_description:
            return False

        if self._how_to_acknowledge != other.acknowledgement:
            return False

        if self._directory_path != other.directory_path:
            return False

        if self._simulations != other.simulations:
            return False

        if self._thumbnail_img != other.thumbnail_image:
            return False

        if self._simu_dt_params != other.simu_datatable_params:
            return False

        return True

    def __ne__(self, other):  # Not an implied relationship between "rich comparison" equality methods in Python 2.X
        return not self.__eq__(other)

    @property
    def category(self):
        """
        :class:`~astrophysix.simdm.ProjectCategory` or
        :attr:`ProjectCategory.alias <astrophysix.simdm.ProjectCategory.alias>` (:obj:`string`). Can be edited."""
        return self._category

    @category.setter
    def category(self, new_cat):
        try:
            scat = Stringifiable.cast_string(new_cat)
            self._category = ProjectCategory.from_alias(scat)
        except ValueError as  ve:
            log.error(str(ve))
            raise AttributeError(str(ve))
        except TypeError:
            if not isinstance(new_cat, ProjectCategory):
                err_msg = "Project 'category' attribute is not a valid ProjectCategory enum value."
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._category = new_cat

    @property
    def project_title(self):
        """Project title"""
        return self._title

    @project_title.setter
    def project_title(self, new_title):
        try:
            self._title = Stringifiable.cast_string(new_title, valid_empty=False)
        except TypeError:
            err_msg = "Project 'project_title' property is not a valid (non empty) string."
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def alias(self):
        """Project alias"""
        return self._alias

    @alias.setter
    def alias(self, new_alias):
        try:
            self._alias = Stringifiable.cast_string(new_alias)
        except TypeError:
            err_msg = "Project 'alias' property is not a valid string"
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def short_description(self):
        """Short description of the project"""
        return self._short_description

    @short_description.setter
    def short_description(self, new_descr):
        try:
            self._short_description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "Project 'short_description' property is not a valid string"
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def general_description(self):
        """General description of the project"""
        return self._general_description

    @general_description.setter
    def general_description(self, new_descr):
        try:
            self._general_description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "Project 'general_description' property is not a valid string"
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def data_description(self):
        """Data description available in this project"""
        return self._data_description

    @data_description.setter
    def data_description(self, new_descr):
        try:
            self._data_description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "Project 'data_description' property is not a valid string"
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def acknowledgement(self):
        """How to acknowledge this project.

        *New in version 0.5.0*.
        """
        return self._how_to_acknowledge

    @acknowledgement.setter
    def acknowledgement(self, ack):
        try:
            self._how_to_acknowledge = Stringifiable.cast_string(ack)
        except TypeError:
            err_msg = "Project 'acknowledgement' property is not a valid string"
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def directory_path(self):
        """Project data directory path"""
        return self._directory_path

    @directory_path.setter
    def directory_path(self, new_path):
        try:
            self._directory_path = Stringifiable.cast_string(new_path)
        except TypeError:
            err_msg = "Project 'directory_path' property is not a valid string"
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def thumbnail_image(self):
        """Thumbnail image for the project"""
        return self._thumbnail_img

    @thumbnail_image.setter
    def thumbnail_image(self, new_thimg):
        """Set new thumbnail image for the current project (for display in the project thumbnail panel on the Galactica
        project category listing

        *New in version 0.7.0*."""
        if new_thimg is None:
            self._thumbnail_img = None
        elif Stringifiable.is_type_string(new_thimg):
            ftype = FileUtil.get_file_type(new_thimg)
            img_file_types = [img_file_class.FILE_TYPE for img_file_class in ImageFile.__subclasses__()]
            if ftype not in img_file_types:
                img_ftypes_str = "/".join([ftype.alias for ftype in img_file_types])
                err_msg = "Only image ({ift:s}) image file paths can be set in {cname:s} as thumbnail " \
                          "images.".format(cname=self.__class__.__name__, kname=ImageFile.__name__, ift=img_ftypes_str)
                log.error(err_msg)
                raise ValueError(err_msg)
            if ftype == FileType.JPEG_FILE:
                self.thumbnail_image = JpegImageFile.load_file(new_thimg)
            elif ftype == FileType.PNG_FILE:
                self._thumbnail_img = PngImageFile.load_file(new_thimg)
            else:
                raise NotImplementedError("Only JPEG and PNG files are accepted as project thumbnail images.")
        elif isinstance(new_thimg, ImageFile):
            self._thumbnail_img = new_thimg
        else:
            err_msg = "Only image (JPEG/PNG) image file paths or {kname:s} objects can be set in {cname:s} as " \
                      "thumbnail images.".format(cname=self.__class__.__name__, kname=ImageFile.__name__)
            log.error(err_msg)
            raise ValueError(err_msg)
        return

    @property
    def simulations(self):
        """Project :class:`~astrophysix.simdm.experiment.Simulation` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._simulations

    @property
    def simu_datatable_params(self):
        """List of :class:`~astrophysix.simdm.protocol.input_parameters.InputParameter` objects coming from the
        project :class:`~astrophysix.simdm.experiment.Simulation` associated :class:`~astrophysix.simdm.protocol.base.SimulationCode`.
        These input parameters can be used in parametric studies to display the project simulations in a datatable
        (instead of a list) on the Galactica project page, (:class:`~astrophysix.simdm.utils.ObjectList`)

        *New in version 0.7.0*.

        Example
        -------
        >>> proj = Project(category="COSMOLOGY", project_title="Reionisation of the universe",
        ...                alias="REIO_4")
        >>> ramses = protocol.SimulationCode(name="Ramses 3.0 (Hydro)", code_name="RAMSES", alias="RAMSES_CODE")
        >>> # Add input parameters
        >>> dx_min = InputParameter(key="max_res", name="dx_min",
        ...                         description="min. spatial resolution")
        >>> ramses.input_parameters.add(dx_min)
        >>> used_mhd_solver = InputParameter(key="with_mhd", name="MHD solver used",
        ...                                  description="Pure hydro or MHD run ?")
        >>> ramses.input_parameters.add(used_mhd_solver)
        >>>
        >>> # Define parameter settings
        >>> dx_min_lores = ParameterSetting(input_param=dx_min, value=100.0, unit=U.kpc)
        >>> dx_min_hires = ParameterSetting(input_param=dx_min, value=10.0, unit=U.kpc)
        >>> use_hydro = ParameterSetting(input_param=used_mhd_solver, value=False)
        >>> use_mhd = ParameterSetting(input_param=used_mhd_solver, value=True)
        >>>
        >>> # Add simulations with defined parameter settings in project
        >>> simu_hydro_lores = experiment.Simulation(ramses, name="My simu low-res (hydro)",
        ...                                          alias="SIMU_HYDRO_LOW", execution_time=exe_time)
        >>> simu_hydro_lores.parameter_settings.add(dx_min_lores)
        >>> simu_hydro_lores.parameter_settings.add(use_hydro)
        >>> proj.simulations.add(simu_hydro_lores)
        >>>
        >>> simu_hydro_hires = experiment.Simulation(ramses, name="My simu high-res (hydro)",
        ...                                          alias="SIMU_HYDRO_HIGH", execution_time=exe_time))
        >>> simu_hydro_hires.parameter_settings.add(dx_min_hires)
        >>> simu_hydro_hires.parameter_settings.add(use_hydro)
        >>> proj.simulations.add(simu_hydro_hires)
        >>>
        >>> simu_mhd_lores = experiment.Simulation(ramses, name="My simu low-res (MHD)",
        ...                                        alias="SIMU_MHD_LOW", execution_time=exe_time)
        >>> simu_mhd_lores.parameter_settings.add(dx_min_lores)
        >>> simu_mhd_lores.parameter_settings.add(use_mhd)
        >>> proj.simulations.add(simu_mhd_lores)
        >>>
        >>> simu_mhd_hires = experiment.Simulation(ramses, name="My simu high-res (MHD)",
        ...                                        alias="SIMU_MHD_HIGH", execution_time=exe_time)
        >>> simu_mhd_hires.parameter_settings.add(dx_min_hires)
        >>> simu_mhd_hires.parameter_settings.add(use_mhd)
        >>> proj.simulations.add(simu_mhd_hires)
        >>>
        >>> # Add input parameters in project simulation datatable parameter list
        >>> proj.simu_datatable_params.add(dx_min)
        >>> proj.simu_datatable_params.add(used_mhd_solver)
        """
        return self._simu_dt_params

    def _simu_codes(self):
        """Simulation codes iterator"""
        puid_list = []
        for simu in self._simulations:
            p = simu.simulation_code
            if p.uid not in puid_list:
                puid_list.append(p.uid)
                yield p

    def _post_pro_codes(self):
        """Post-processing codes iterator"""
        ppcode_uid_list = []
        for simu in self._simulations:
            for pprun in simu.post_processing_runs:
                p = pprun.postpro_code
                if p.uid not in ppcode_uid_list:
                    ppcode_uid_list.append(p.uid)
                    yield p

    def _target_objects(self):
        """Target object iterator"""
        to_uid_list = []
        for simu in self._simulations:  # Loop over simulations
            for sn in simu.snapshots:  # Loop over simulation snapshots
                for cat in sn.catalogs:  # Loop over catalogs
                    if cat.target_object.uid not in to_uid_list:
                        to_uid_list.append(cat.target_object.uid)
                        yield cat.target_object
            for res in simu.generic_results:  # Loop over simulation generic results
                for cat in res.catalogs:  # Loop over catalogs
                    if cat.target_object.uid not in to_uid_list:
                        to_uid_list.append(cat.target_object.uid)
                        yield cat.target_object

            for pprun in simu.post_processing_runs:
                for sn in pprun.snapshots:  # Loop over post-processing run snapshots
                    for cat in sn.catalogs:  # Loop over catalogs
                        if cat.target_object.uid not in to_uid_list:
                            to_uid_list.append(cat.target_object.uid)
                            yield cat.target_object
                for res in pprun.generic_results:  # Loop over post-processing run generic results
                    for cat in res.catalogs:  # Loop over catalogs
                        if cat.target_object.uid not in to_uid_list:
                            to_uid_list.append(cat.target_object.uid)
                            yield cat.target_object

    def _check_valid_simu_input_param(self, input_param):
        """
        Checks that a given input parameter can be added into this project's aimulation datatable input parameter list.
        Verifies that the input parameter belongs to any of the project experiment protocol's input parameter list.
        Raises an AttributeError if not.

        Parameters
        ----------
        input_param: ``astrophysix.simdm.protocol.input_parameter.InputParameter``
            input parameter to add into the simulation datatable parameter list
        """
        for simu_code in self._simu_codes():
            if input_param not in simu_code.input_parameters:
                err_msg = "{cname:s} '{ip!s}' does not refer to any input parameter of the project's simulation " \
                          "codes.".format(cname=self.__class__.__name__, ip=input_param)
                log.error(err_msg)
                raise AttributeError(err_msg)

    def _can_delete_input_param(self, inp_param):
        """
        Checks if an input parameter does not belong to the project simulation's datatable parameter list and can be
        safely deleted. Returns None if it can be deleted, otherwise returns a string.

        Parameters
        ----------
        inp_param: ``astrophysix.simdm.protocol.input_parameters.InputParameter``
            input parameter about to be deleted

        Returns
        -------
        o: str or None
        """
        for ip in self._simu_dt_params:
            if ip is inp_param:  # Reference identity, not equality ??? Should work
                return "Simulation datatable {ip!s}".format(s=self, ip=inp_param)
        return None

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["category", "project_title", "alias", "short_description", "general_description", "data_description",
                "data_description", "acknowledgement", "directory_path", "thumbnail_image"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a Project object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the Project into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(Project, self)._hsp_write(h5group, **kwargs)

        # If necessary, call callback function with project name
        self._hsp_write_callback(str(self), **kwargs)

        # Write project title
        self._hsp_write_attribute(h5group, ('title', self._title), **kwargs)

        # Write project category alias
        self._hsp_write_attribute(h5group, ('category', self._category.alias), **kwargs)

        # Write project Galactica alias, if defined
        self._hsp_write_attribute(h5group, ('galactica_alias', self._alias), **kwargs)

        # Write project directory path, if defined
        self._hsp_write_attribute(h5group, ('project_directory', self._directory_path), **kwargs)

        # Write project short/general/data description/ acknowledgement
        self._hsp_write_attribute(h5group, ('short_description', self._short_description), **kwargs)
        self._hsp_write_attribute(h5group, ('general_description', self._general_description), **kwargs)
        self._hsp_write_attribute(h5group, ('data_description', self._data_description), **kwargs)
        self._hsp_write_attribute(h5group, ('acknowledgement', self._how_to_acknowledge), **kwargs)

        # Write project thumbnail image into HDF5 file
        if not kwargs.get("dry_run", False) and not kwargs.get("new_file", True):  # Old HDF5 file being modified (not a dry run)
            # Delete thumbnail image from HDF5 group if not present anymore in the Project
            if "thumbnail_img" in h5group and self.thumbnail_image is None:
                del h5group["thumbnail_img"]

        self._hsp_write_object(h5group, "thumbnail_img", self._thumbnail_img, **kwargs)

        # Write protocol directory
        if kwargs.get("from_project", False):  # Write protocol list in project subgroup (not in each experiment)
            proto_group = self._hsp_get_or_create_h5group(h5group, "PROTOCOLS", **kwargs)
            self._hsp_write_object_list(proto_group, "SIMU_CODES", self._simu_codes, "simu_code_", **kwargs)
            self._hsp_write_object_list(proto_group, "PPRUN_CODES", self._post_pro_codes, "pprun_code_", **kwargs)

        # Write all target objects
        if kwargs.get("from_project", False): # Write target object list in project subgroup (not in each catalog)
            self._hsp_write_object_list(h5group, "TARGET_OBJECTS", self._target_objects, "targobj_", **kwargs)

        # Write all simulations
        self._hsp_write_object_list(h5group, "SIMULATIONS", self._simulations, "simu_", **kwargs)

        # Write simulation datatable parameters
        self._hsp_write_object_list(h5group, "SIMU_DT_PARAMS", self._simu_dt_params, "simu_dt_param_", **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a Project object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        proj: ``Project``
            Read Project instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(Project, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Create Project instance with mandatory attributes
        t = cls._hsp_read_attribute(h5group, "title", "project title")
        cat = cls._hsp_read_attribute(h5group, "category", "project category")
        proj = cls(uid=uid, category=cat, project_title=t)

        # ------------------------------------------ Optional attributes --------------------------------------------- #
        # Read Galactica alias
        alias = cls._hsp_read_attribute(h5group, "galactica_alias", "project Galactica alias",
                                        raise_error_if_not_found=False)
        if alias is  not None:
            proj.alias = alias
        # Read project directory
        dpath = cls._hsp_read_attribute(h5group, "project_directory", "project directory",
                                        raise_error_if_not_found=False)
        if dpath is not None:
            proj.directory_path = dpath

        # -------------------------- Read project short/general/data description --------------------------- #
        ddescr = cls._hsp_read_attribute(h5group, "data_description", "project data description",
                                         raise_error_if_not_found=False)
        if ddescr is not None:
            proj.data_description = ddescr
        gdescr = cls._hsp_read_attribute(h5group, "general_description", "project general description",
                                         raise_error_if_not_found=False)
        if gdescr is not None:
            proj.general_description = gdescr

        sdescr = cls._hsp_read_attribute(h5group, "short_description", "project short description",
                                         raise_error_if_not_found=False)
        if sdescr is not None:
            proj.short_description = sdescr
        # --------------------------------------------------------------------------------------------------- #
        # ------------------------------------------------------------------------------------------------------------ #

        # Build dependency object dictionary indexed by their class name
        dod = {}
        if "PROTOCOLS" in h5group:
            protgroup = h5group["PROTOCOLS"]
            # Build simulation code dictionary indexed by their UUID
            if "SIMU_CODES" in protgroup:
                simu_code_dict = {}
                for simu_code in SimulationCode._hsp_read_object_list(protgroup, "SIMU_CODES", "simu_code_",
                                                                      "simulation code"):
                    simu_code_dict[simu_code.uid] = simu_code

                dod[SimulationCode.__name__] = simu_code_dict

            # Build post-processing code dictionary indexed by their UUID
            if "PPRUN_CODES" in protgroup:
                pprun_code_dict = {}
                for pprun_code in PostProcessingCode._hsp_read_object_list(protgroup, "PPRUN_CODES", "pprun_code_",
                                                                           "post-processing code"):
                    pprun_code_dict[pprun_code.uid] = pprun_code

                dod[PostProcessingCode.__name__] = pprun_code_dict

        if version >= 2:  # and "TARGET_OBJECTS" in h5group:
            # Build target object dictionary indexed by their UUID
            tobj_dict = {}
            for tobj in TargetObject._hsp_read_object_list(h5group, "TARGET_OBJECTS", "targobj_", "target object",
                                                           dependency_objdict=dod):
                tobj_dict[tobj.uid] = tobj
            dod[TargetObject.__name__] = tobj_dict

            # Read project acknowledgement text
            ackn = cls._hsp_read_attribute(h5group, "acknowledgement", "project acknowledgement",
                                           raise_error_if_not_found=False)
            if ackn is not None:
                proj.acknowledgement = ackn

        if version >= 3:
            # Read project thumbnail image file
            if "thumbnail_img" in h5group:
                proj.thumbnail_image = ImageFile._hsp_read_object(h5group, "thumbnail_img", "thumbnail image file")

        # Build simulation list and add each simulation into project
        if "SIMULATIONS" in h5group:
            for simu in Simulation._hsp_read_object_list(h5group, "SIMULATIONS", "simu_", "project simulation",
                                                         dependency_objdict=dod):
                proj.simulations.add(simu)

        if version >= 4:
            if "SIMU_DT_PARAMS" in h5group:
                for simudt_param in InputParameter._hsp_read_object_list(h5group, "SIMU_DT_PARAMS", "simu_dt_param_",
                                                                         "project's simulation datatable parameter",
                                                                         dependency_objdict=dod):
                    proj.simu_datatable_params.add(simudt_param)

        return proj

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: `dict`
            keyword arguments (optional)
        """
        # Check project alias
        if len(self.alias) == 0:
            log.warning("{p!s} Galactica alias is missing.".format(p=self))
        elif len(self._alias) > 16:
            log.warning("{p!s} Galactica alias is too long (max. 16 characters).".format(p=self))
        else:
            err_msg = self.galactica_valid_alias(self._alias)
            if err_msg is not None:
                log.warning("{p!s} Galactica alias is not valid ({m:s})".format(p=self, m=err_msg))

        # Check project title
        if len(self._title) == 0:
            log.warning("{p!s} Galactica project title is missing.".format(p=self))
        elif len(self._title) > 128:
            log.warning("{p!s} Galactica project title is too long (max. 128 characters).".format(p=self))

        # Check project short description
        if len(self._short_description) == 0:
            log.warning("{p!s} Galactica short description is missing.".format(p=self))
        elif len(self._short_description) > 256:
            log.warning("{p!s} Galactica short description is too long (max. 256 characters).".format(p=self))

        # Check project post-processing/simulation code validity + protocol alias unicity
        code_names = {}
        for scode in self._simu_codes():
            scode.galactica_validity_check()
            if len(scode.alias) > 0:
                if scode.alias in code_names:
                    log.warning("{c1!s} and {c2!s} protocols share the same alias. They must be "
                                "unique.".format(c1=code_names[scode.alias], c2=scode))
                else:
                    code_names[scode.alias] = scode
        for pcode in self._post_pro_codes():
            pcode.galactica_validity_check()
            if len(pcode.alias) > 0:
                if pcode.alias in code_names:
                    log.warning("{c1!s} and {c2!s} protocols share the same alias. They must be "
                                "unique.".format(c1=code_names[pcode.alias], c2=pcode))
                else:
                    code_names[pcode.alias] = pcode

        # Check project post-processing runs/simulations validity + experiment alias unicity
        experiments = {}
        for srun in self._simulations:
            srun.galactica_validity_check()
            if len(srun.alias) > 0:
                if srun.alias in experiments:
                    log.warning("{r1!s} and {r2!s} experiments share the same alias. They must be "
                                "unique.".format(r1=experiments[srun.alias], r2=srun))
                else:
                    experiments[srun.alias] = srun

            for prun in srun.post_processing_runs:
                prun.galactica_validity_check()
                if len(prun.alias) > 0:
                    if prun.alias in experiments:
                        log.warning("{r1!s} and {r2!s} experiments share the same alias. They must be "
                                    "unique.".format(r1=experiments[prun.alias], r2=prun))
                    else:
                        experiments[prun.alias] = prun

        # Check target object validity + object name unicity in the project
        targobj_names = {}
        for targobj in self._target_objects():
            targobj.galactica_validity_check()
            if targobj.name in targobj_names:
                log.warning("{r1!s} and {r2!s} share the same name. They must be "
                            "unique.".format(r1=targobj_names[targobj.name], r2=targobj))
            else:
                targobj_names[targobj.name] = targobj

    def __unicode__(self):
        """
        String representation of the instance
        """
        strrep = "[{category:s}]".format(category=self._category.verbose_name)

        # Title and short description
        if len(self._title) > 0:
            strrep += " '{ptitle:s}' project".format(ptitle=self._title)

        return strrep


__all__ = ["Project", "ProjectCategory"]
